import pyglet
import random
import math
from . import resources


def distance(point_1=(0, 0), point_2=(0, 0)):
    """Returns the distance between two points"""
    return math.sqrt((point_1[0] - point_2[0]) ** 2 + (point_1[1] - point_2[1]) ** 2)


def asteroids(num_asteroids, player_position):
    """Generate asteroid objects with random positions and velocities, 
    not close to the player"""
    asteroids = []
    for i in range(num_asteroids):
        asteroid_x, asteroid_y, _ = player_position
        while distance((asteroid_x, asteroid_y), player_position) < 100:
            asteroid_x = random.randint(0, 800)
            asteroid_y = random.randint(0, 600)
        new_asteroid = pyglet.sprite.Sprite(img=resources.asteroid_image,
                                            x=asteroid_x, y=asteroid_y)
        new_asteroid.rotation = random.randint(0, 360)
        asteroids.append(new_asteroid)
    return asteroids



"""Plain text decoder."""
from __future__ import annotations

from typing import TYPE_CHECKING

import pyglet

if TYPE_CHECKING:
    from pyglet.resource import Location
    from pyglet.text import UnformattedDocument


class PlainTextDecoder(pyglet.text.DocumentDecoder):  # noqa: D101
    def decode(self, text: str, location: Location | None=None) -> UnformattedDocument:  # noqa: ARG002
        document = pyglet.text.document.UnformattedDocument()
        document.insert_text(0, text)
        return document



"""
Simple example showing some animated shapes
"""
import math
import pyglet
from pyglet import shapes


class ShapesDemo(pyglet.window.Window):

    def __init__(self, width, height):
        super().__init__(width, height, "Shapes")
        self.time = 0
        self.batch = pyglet.graphics.Batch()

        self.circle = shapes.Circle(360, 240, 75, color=(255, 225, 255, 127), batch=self.batch)

        # Rectangle with center as anchor
        self.square = shapes.BorderedRectangle(360, 240, 100, 100, border=5, color=(55, 55, 255),
                                               border_color=(25, 25, 25), batch=self.batch)
        self.square.anchor_position = 50, 50

        # Large transparent rectangle
        self.rectangle = shapes.Rectangle(100, 190, 500, 100, color=(255, 22, 20, 64), batch=self.batch)

        self.line = shapes.Line(0, 0, 0, 480, thickness=4, color=(200, 20, 20), batch=self.batch)

        self.triangle = shapes.Triangle(10, 10, 190, 10, 100, 150, color=(55, 255, 255, 175), batch=self.batch)

        septagon_step = math.pi * 2 / 7
        self.fading_septagon = shapes.Polygon(
            *[[50 + 40 * math.sin(i * septagon_step), 200 + 40 * math.cos(i * septagon_step)] for i in range(7)],
            batch=self.batch,
        )

        self.arc = shapes.Arc(50, 300, radius=40, segments=25, angle=270.0, color=(255, 255, 255), batch=self.batch)

        self.star = shapes.Star(600, 375, 50, 30, 5, color=(255, 255, 0), batch=self.batch)

        self.ellipse = shapes.Ellipse(650, 150, a=50, b=30, color=(55, 255, 55), batch=self.batch)

        self.sector = shapes.Sector(125, 400, 60, angle=0.45 * 360, color=(55, 255, 55), batch=self.batch)

        self.polygon = shapes.Polygon([400, 100], [500, 10], [600, 100], [550, 175], [450, 150], batch=self.batch)

        self.box = shapes.Box(60, 40, 200, 100, thickness=2, color=(244, 55, 55), batch=self.batch)

        coordinates = [[450, 400], [475, 450], [525, 450], [550, 400]]
        self.multiLine = shapes.MultiLine(*coordinates, closed=True, batch=self.batch)

    def on_draw(self):
        """Clear the screen and draw shapes"""
        self.clear()
        self.batch.draw()

    def update(self, delta_time):
        """Animate the shapes"""
        self.time += delta_time
        self.square.rotation = self.time * 15
        self.rectangle.y = 200 + math.sin(self.time) * 190
        self.circle.radius = 75 + math.sin(self.time * 1.17) * 25
        self.triangle.rotation = self.time * 15

        self.line.x = 360 + math.sin(self.time * 0.81) * 360
        self.line.x2 = 360 + math.sin(self.time * 1.34) * 360

        self.arc.rotation = self.time * 30

        self.fading_septagon.opacity = int(255 * (0.5 + (0.5 * math.cos(self.time))))

        self.star.rotation = self.time * 50
        self.polygon.rotation = self.time * 45

        self.ellipse.b = abs(math.sin(self.time) * 100)
        self.sector.angle = (self.time * 30) % 360.0

        self.multiLine.rotation = self.time * -15

        self.multiLine.rotation = self.time * -15


if __name__ == "__main__":
    demo = ShapesDemo(720, 480)
    pyglet.clock.schedule_interval(demo.update, 1/30)
    pyglet.app.run()



"""Demonstrates creation of a transparent overlay window in pyglet
"""

import pyglet

from pyglet.graphics import Batch
from pyglet.window import Window


batch = Batch()
window = Window(500, 500, style=Window.WINDOW_STYLE_TRANSPARENT)
window.set_caption("Transparent Window")

circle = pyglet.shapes.Circle(250, 250, 100, color=(255, 255, 0), batch=batch)


@window.event
def on_draw():
    window.clear()
    batch.draw()


pyglet.app.run()



from __future__ import annotations

from functools import partial
from unittest.mock import MagicMock

import pytest

from pyglet.graphics import Batch, Group
from pyglet.shapes import *


# The shapes are tested individually since their RGBA handling is
# inlined for maximum speed instead of encapsulated in their baseclass.
# A typo might break color functionality in one but not the others.
@pytest.fixture(scope="module", params=[
    (Arc, (0, 0, 5)),
    (Circle, (0, 0, 5)),
    # Ellipse's a value below is nonsensical in normal use, but here it
    # makes sure the value is not confused with the RGBA alpha channel
    # internally.
    (Ellipse, (0, 0, 0, 5)),
    (Sector, (0, 0, 3)),
    (Line, (0, 0, 7, 7)),
    (Rectangle, (0, 0, 20, 20)),
    (BorderedRectangle, (0, 0, 30, 10)),
    (Triangle, (0, 0, 2, 2, 5, 5)),
    (Star, (1, 1, 20, 11, 5)),
    (Polygon, ((0, 0), (1, 1), (2, 2))),
    (MultiLine, ((0, 0), (1, 1), (2, 2))),
])
def shape_and_positionals(request):
    return request.param


# Enable type-specific behavior; just Line + rotation at the moment
@pytest.fixture()
def shape_type(shape_and_positionals):
    return shape_and_positionals[0]


@pytest.fixture()
def shape_keywords_only(shape_and_positionals):
    class_, positional_args = shape_and_positionals
    return partial(class_, *positional_args)


@pytest.fixture()
def rgb_or_rgba_shape(shape_keywords_only, original_rgb_or_rgba_color):
    return shape_keywords_only(color=original_rgb_or_rgba_color)


@pytest.fixture()
def rgba_shape(shape_keywords_only):
    return shape_keywords_only(color=(0, 255, 0, 37))


def test_init_sets_opacity_from_rgba_value_as_color_argument(rgba_shape):
    assert rgba_shape.opacity == 37


def test_init_sets_opacity_to_255_for_rgb_value_as_color_argument(shape_keywords_only):
    assert shape_keywords_only(color=(0, 0, 0)).opacity == 255


def test_init_sets_rotation_to_zero(rgb_or_rgba_shape, shape_type):
    if shape_type is Line:
        pytest.xfail("Rotation test not yet valid for line due to design ambiguity")
    assert rgb_or_rgba_shape.rotation == 0


def test_rotation_prop_sets_rotation(rgb_or_rgba_shape, new_nonzero_rotation):
    rgb_or_rgba_shape.rotation = new_nonzero_rotation
    assert rgb_or_rgba_shape.rotation == new_nonzero_rotation


def test_setting_color_sets_color_rgb_channels(rgb_or_rgba_shape, new_rgb_or_rgba_color):
    rgb_or_rgba_shape.color = new_rgb_or_rgba_color
    assert rgb_or_rgba_shape.color[:3] == new_rgb_or_rgba_color[:3]


def test_setting_color_to_rgb_value_does_not_change_opacity(rgb_or_rgba_shape, new_rgb_color):
    original_opacity = rgb_or_rgba_shape.opacity
    rgb_or_rgba_shape.color = new_rgb_color
    assert rgb_or_rgba_shape.opacity == original_opacity


def test_setting_color_to_rgba_value_changes_opacity(rgb_or_rgba_shape, new_rgba_color):
    rgb_or_rgba_shape.color = new_rgba_color
    assert rgb_or_rgba_shape.opacity == new_rgba_color[3]
    assert rgb_or_rgba_shape.color[3] == new_rgba_color[3]


def test_setting_opacity_does_not_change_rgb_channels_on_color(rgb_or_rgba_shape):
    original_color = rgb_or_rgba_shape.color[:3]
    rgb_or_rgba_shape.opacity = 255
    assert rgb_or_rgba_shape.color[:3] == original_color


def test_group_setter(shape_keywords_only):
    shape = shape_keywords_only()

    new_group = Group()
    shape.group = new_group
    assert shape.group is new_group


def test_batch_setter(shape_keywords_only):
    shape = shape_keywords_only()

    new_batch = Batch()
    shape.batch = new_batch
    assert shape.batch is new_batch


def test_program_setter(shape_keywords_only):
    shape = shape_keywords_only()

    program = MagicMock()
    shape.program = program
    assert shape.program == program


def test_blend_setter(shape_keywords_only):
    shape = shape_keywords_only()

    blend_mode = (1, 1)
    shape.blend_mode = blend_mode
    assert shape._group.blend_src == 1  # noqa: SLF001
    assert shape._group.blend_dest == 1  # noqa: SLF001



from abc import ABCMeta, abstractmethod


class AbstractListener(metaclass=ABCMeta):
    """The listener properties for positional audio.

    You can obtain the singleton instance of this class by calling
    :meth:`AbstractAudioDriver.get_listener`.
    """

    _volume = 1.0
    _position = (0, 0, 0)
    _forward_orientation = (0, 0, -1)
    _up_orientation = (0, 1, 0)

    @abstractmethod
    def _set_volume(self, volume):
        pass

    volume = property(lambda self: self._volume,
                      lambda self, volume: self._set_volume(volume),
                      doc="""The master volume for sound playback.

        All sound volumes are multiplied by this master volume before being
        played.  A value of 0 will silence playback (but still consume
        resources).  The nominal volume is 1.0.

        :type: float
        """)

    @abstractmethod
    def _set_position(self, position):
        pass

    position = property(lambda self: self._position,
                        lambda self, position: self._set_position(position),
                        doc="""The position of the listener in 3D space.

        The position is given as a tuple of floats (x, y, z).  The unit
        defaults to meters, but can be modified with the listener
        properties.

        :type: 3-tuple of float
        """)

    @abstractmethod
    def _set_forward_orientation(self, orientation):
        pass

    forward_orientation = property(lambda self: self._forward_orientation,
                               lambda self, o: self._set_forward_orientation(o),
                               doc="""A vector giving the direction the
        listener is facing.

        The orientation is given as a tuple of floats (x, y, z), and has
        no unit.  The forward orientation should be orthogonal to the
        up orientation.

        :type: 3-tuple of float
        """)

    @abstractmethod
    def _set_up_orientation(self, orientation):
        pass

    up_orientation = property(lambda self: self._up_orientation,
                              lambda self, o: self._set_up_orientation(o),
                              doc="""A vector giving the "up" orientation
        of the listener.

        The orientation is given as a tuple of floats (x, y, z), and has
        no unit.  The up orientation should be orthogonal to the
        forward orientation.

        :type: 3-tuple of float
        """)




"""Collection of image encoders and decoders.

Modules must subclass ImageDecoder and ImageEncoder for each method of
decoding/encoding they support.

Modules must also implement the two functions::

    def get_decoders():
        # Return a list of ImageDecoder instances or []
        return []

    def get_encoders():
        # Return a list of ImageEncoder instances or []
        return []

"""
from __future__ import annotations

import os.path

from pyglet.util import CodecRegistry, Decoder, Encoder, DecodeException, EncodeException
from pyglet import compat_platform


class _ImageCodecRegistry(CodecRegistry):
    """Subclass of CodecRegistry that adds support for animation methods."""

    def __init__(self):
        self._decoder_animation_extensions = {}
        super().__init__()

    def add_decoders(self, module):
        """Override the default method to also add animation decoders."""
        super().add_decoders(module)
        for decoder in module.get_decoders():
            for extension in decoder.get_animation_file_extensions():
                if extension not in self._decoder_animation_extensions:
                    self._decoder_animation_extensions[extension] = []
                self._decoder_animation_extensions[extension].append(decoder)

    def get_animation_decoders(self, filename=None):
        """Get a list of animation decoders. If a `filename` is provided, only
           decoders supporting that extension will be returned. An empty list
           will be return if no encoders for that extension are available.
        """
        if filename:
            extension = os.path.splitext(filename)[1].lower()
            return self._decoder_animation_extensions.get(extension, [])
        return self._decoders

    def decode_animation(self, filename, file, **kwargs):
        first_exception = None

        for decoder in self.get_animation_decoders(filename):
            try:
                return decoder.decode_animation(filename, file, **kwargs)
            except DecodeException as e:
                if not first_exception:
                    first_exception = e
                if file:
                    file.seek(0)

        for decoder in self.get_animation_decoders():   # Try ALL codecs
            try:
                return decoder.decode_animation(filename, file, **kwargs)
            except DecodeException:
                if file:
                    file.seek(0)

        if not first_exception:
            msg = f"No decoders available for this file type: {filename}"
            raise DecodeException(msg)
        raise first_exception


registry = _ImageCodecRegistry()
add_decoders = registry.add_decoders
add_encoders = registry.add_encoders
get_animation_decoders = registry.get_animation_decoders
get_decoders = registry.get_decoders
get_encoders = registry.get_encoders


class ImageDecodeException(DecodeException):
    pass


class ImageEncodeException(EncodeException):
    pass


class ImageDecoder(Decoder):

    def get_animation_file_extensions(self) -> list[str]:
        """Return a list of accepted file extensions, e.g. ['.gif', '.flc']
        Lower-case only.
        """
        return []

    def decode(self, filename, file):
        """Decode the given file object and return an instance of `Image`.
        Throws ImageDecodeException if there is an error.  filename
        can be a file type hint.
        """
        raise NotImplementedError()

    def decode_animation(self, filename, file):
        """Decode the given file object and return an instance of :py:class:`~pyglet.image.Animation`.
        Throws ImageDecodeException if there is an error.  filename
        can be a file type hint.
        """
        raise ImageDecodeException('This decoder cannot decode animations.')

    def __repr__(self) -> str:
        return "{}{}".format(self.__class__.__name__,
                               self.get_animation_file_extensions() +
                               self.get_file_extensions())


class ImageEncoder(Encoder):

    def encode(self, image, filename, file):
        """Encode the given image to the given file.  filename
        provides a hint to the file format desired.
        """
        raise NotImplementedError()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}{self.get_file_extensions()}"


def add_default_codecs():
    # Add the codecs we know about.  These should be listed in order of
    # preference.  This is called automatically by pyglet.image.

    # Compressed texture in DDS format
    try:
        from pyglet.image.codecs import dds
        registry.add_encoders(dds)
        registry.add_decoders(dds)
    except ImportError:
        pass

    # Mac OS X default: Quartz
    if compat_platform == 'darwin':
        try:
            from pyglet.image.codecs import quartz
            registry.add_encoders(quartz)
            registry.add_decoders(quartz)
        except ImportError:
            pass

    # Windows 7 default: Windows Imaging Component
    if compat_platform in ('win32', 'cygwin'):
        from pyglet.libs.win32.constants import WINDOWS_7_OR_GREATER
        if WINDOWS_7_OR_GREATER:  # Supports Vista and above.
            try:
                from pyglet.image.codecs import wic
                registry.add_encoders(wic)
                registry.add_decoders(wic)
            except ImportError:
                pass

    # Windows XP default: GDI+
    if compat_platform in ('win32', 'cygwin'):
        try:
            from pyglet.image.codecs import gdiplus
            registry.add_encoders(gdiplus)
            registry.add_decoders(gdiplus)
        except ImportError:
            pass

    # Linux default: GdkPixbuf 2.0
    if compat_platform.startswith('linux'):
        try:
            from pyglet.image.codecs import gdkpixbuf2
            registry.add_encoders(gdkpixbuf2)
            registry.add_decoders(gdkpixbuf2)
        except ImportError:
            pass

    # Fallback: PIL
    try:
        from pyglet.image.codecs import pil
        registry.add_encoders(pil)
        registry.add_decoders(pil)
    except ImportError:
        pass

    # Fallback: PNG loader (slow)
    try:
        from pyglet.image.codecs import png
        registry.add_encoders(png)
        registry.add_decoders(png)
    except ImportError:
        pass

    # Fallback: BMP loader (slow)
    try:
        from pyglet.image.codecs import bmp
        registry.add_encoders(bmp)
        registry.add_decoders(bmp)
    except ImportError:
        pass



import unittest

from pyglet.image import atlas

__noninteractive = True


class Rect:
    def __init__(self, x1, y1, x2, y2):
        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2

    def __repr__(self):
        return 'Rect(%d, %d to %d, %d)' % (
            self.x1, self.y1, self.x2, self.y2)

    def intersects(self, other):
        return self.x2 > other.x1 and self.x1 < other.x2 and \
               self.y2 > other.y1 and self.y1 < other.y2


class AllocatorEnvironment:
    def __init__(self, test_case, width, height):
        self.test_case = test_case
        self.rectes = []
        self.allocator = atlas.Allocator(width, height)

    def check(self, test_case):
        for i, rect in enumerate(self.rectes):
            test_case.assertTrue(0 <= rect.x1 < self.allocator.width)
            test_case.assertTrue(0 <= rect.x2 <= self.allocator.width)
            test_case.assertTrue(0 <= rect.y1 < self.allocator.height)
            test_case.assertTrue(0 <= rect.y2 <= self.allocator.height)
            for other in self.rectes[i + 1:]:
                test_case.assertFalse(rect.intersects(other))

    def add(self, width, height):
        x, y = self.allocator.alloc(width, height)
        self.rectes.append(Rect(x, y, x + width, y + height))
        self.check(self.test_case)

    def add_fail(self, width, height):
        self.test_case.assertRaises(atlas.AllocatorException,
                                    self.allocator.alloc, width, height)


class TestPack(unittest.TestCase):
    def test_over_x(self):
        env = AllocatorEnvironment(self, 3, 3)
        env.add_fail(3, 4)

    def test_over_y(self):
        env = AllocatorEnvironment(self, 3, 3)
        env.add_fail(4, 3)

    def test_1(self):
        env = AllocatorEnvironment(self, 4, 4)
        for i in range(16):
            env.add(1, 1)
        env.add_fail(1, 1)

    def test_2(self):
        env = AllocatorEnvironment(self, 3, 3)
        env.add(2, 2)
        for i in range(4):
            env.add(1, 1)

    def test_3(self):
        env = AllocatorEnvironment(self, 3, 3)
        env.add(3, 3)
        env.add_fail(1, 1)

    def test_4(self):
        env = AllocatorEnvironment(self, 5, 4)
        for i in range(4):
            env.add(2, 2)
        env.add_fail(2, 1)
        env.add(1, 2)
        env.add(1, 2)
        env.add_fail(1, 1)

    def test_5(self):
        env = AllocatorEnvironment(self, 4, 4)
        env.add(3, 2)
        env.add(4, 2)
        env.add(1, 2)
        env.add_fail(1, 1)


if __name__ == '__main__':
    unittest.main()



import os
import array
import shutil
import warnings

import pytest

import pyglet
from pyglet.image import get_buffer_manager

from .data import PygletTestCase

local_dir = os.path.dirname(__file__)
test_dir = os.path.normpath(os.path.join(local_dir, '..'))

base_screenshot_path = os.path.join(test_dir, 'interactive', 'screenshots')
committed_screenshot_path = os.path.join(base_screenshot_path, 'committed')
session_screenshot_path = os.path.join(base_screenshot_path, 'session')

del local_dir, test_dir


class InteractiveFixture:
    """Fixture for interactive test cases. Provides interactive prompts and
    verifying screenshots.
    """
    def __init__(self, request):
        self.screenshots = []
        self._request = request

    @property
    def interactive(self):
        return not self.sanity and not self.non_interactive

    @property
    def sanity(self):
        return self._request.config.getoption('--sanity', False)

    @property
    def non_interactive(self):
        return self._request.config.getoption('--non-interactive', False)

    @property
    def allow_missing_screenshots(self):
        return not self.non_interactive

    @property
    def testname(self):
        parts = []
        parts.append(self._request.node.module.__name__)
        if self._request.node.cls:
            parts.append(self._request.node.cls.__name__)
        parts.append(self._request.node.name)
        return '.'.join(parts)

    def ask_question(self, description=None):
        """Ask a question to verify the current test result. Uses the console or an external gui
        as no window is available."""
        failure_description = None
        if self.interactive:
            failure_description = _ask_user_to_verify(description)
            if failure_description is not None:
                self.fail(failure_description)

    def _take_screenshot(self, window=None):
        """
        Take a screenshot to allow visual verification.
        """
        screenshot_name = self._get_next_screenshot_name()
        screenshot_file_name = self._get_screenshot_session_file_name(screenshot_name)

        if window is not None:
            window.switch_to()

        get_buffer_manager().get_color_buffer().get_image_data().save(screenshot_file_name)
        self.screenshots.append(screenshot_name)
        self._schedule_commit()

        return screenshot_name

    def _check_screenshot(self, screenshot_name):
        session_file_name = self._get_screenshot_session_file_name(screenshot_name)
        committed_file_name = self._get_screenshot_committed_file_name(screenshot_name)

        assert os.path.isfile(session_file_name)
        if os.path.isfile(committed_file_name):
            committed_image = pyglet.image.load(committed_file_name)
            session_image = pyglet.image.load(session_file_name)
            self.assert_image_equal(committed_image, session_image)
        else:
            assert self.allow_missing_screenshots
            warnings.warn('No committed reference screenshot available.')

    def _get_next_screenshot_name(self):
        """
        Get the unique name for the next screenshot.
        """
        return '{}.{:03d}.png'.format(self.testname,
                                      len(self.screenshots)+1)

    def _get_screenshot_session_file_name(self, screenshot_name):
        return os.path.join(session_screenshot_path, screenshot_name)

    def _get_screenshot_committed_file_name(self, screenshot_name):
        return os.path.join(committed_screenshot_path, screenshot_name)

    def _schedule_commit(self):
        if not hasattr(self._request.session, 'pending_screenshots'):
            self._request.session.pending_screenshots = set()
        self._request.session.pending_screenshots.add(self)

    def assert_image_equal(self, a, b, tolerance=0, msg=None):
        if msg is None:
            msg = 'Screenshot does not match last committed screenshot'
        if a is None:
            assert b is None, msg
        else:
            assert b is not None, msg

        a_data = a.get_image_data()
        b_data = b.get_image_data()

        assert a_data.width == b_data.width, msg
        assert a_data.height == b_data.height, msg
        assert a_data.format == b_data.format, msg
        assert a_data.pitch == b_data.pitch, msg
        self.assert_buffer_equal(a_data.data, b_data.data, tolerance, msg)

    def assert_buffer_equal(self, a, b, tolerance=0, msg=None):
        if tolerance == 0:
            assert a == b, msg

        assert len(a) == len(b), msg

        a = array.array('B', a)
        b = array.array('B', b)
        for (aa, bb) in zip(a, b):
            assert abs(aa - bb) <= tolerance, msg

    def commit_screenshots(self):
        """
        Store the screenshots for reference if the test case is successful.
        """
        for screenshot_name in self.screenshots:
            shutil.copyfile(self._get_screenshot_session_file_name(screenshot_name),
                            self._get_screenshot_committed_file_name(screenshot_name))


class InteractiveTestCase(PygletTestCase):
    """
    Base class for interactive tests.

    Interactive tests exist on several levels of interactivity. The least interactive tests store
    screenshots of user validated output that can be used to automatically compare to in a non
    interactive run. More interactive tests cannot validate the results automatically, but can
    run fully automatic for sanity checks. Finally there are tests that really require the user
    to perform an action for the test to continue.

    Use the decorators @only_interactive, @requires_user_validation and @requires_user_action to
    mark a test case as such. This only works on the test suite (class) level.
    """
    # Show interactive prompts
    interactive = True

    # Allow tests missing reference screenshots to pass
    allow_missing_screenshots = False

    def __init__(self, methodName):
        super().__init__(methodName=methodName)
        self._screenshots = []

    def check_screenshots(self):
        # If we arrive here, there have not been any failures yet
        if self.interactive:
            self._commit_screenshots()
        else:
            if self._has_reference_screenshots():
                self._validate_screenshots()
                # Always commit the screenshots here. They can be used for the next test run.
                # If reference screenshots were already present and there was a mismatch, it should
                # have failed above.
                self._commit_screenshots()

            elif self.allow_missing_screenshots:
                warnings.warn('No committed reference screenshots available. Ignoring.')
            else:
                self.fail('No committed reference screenshots available. Run interactive first.')

    def user_verify(self, description, take_screenshot=True):
        """
        Request the user to verify the current display is correct.
        """
        failed = False
        failure_description = None

        if self.interactive:
            failure_description = _ask_user_to_verify(description)
        if take_screenshot:
            self._take_screenshot()

        if failure_description is not None:
            self.fail(failure_description)

    def assert_image_equal(self, a, b, tolerance=0, msg=None):
        if msg is None:
            msg = 'Screenshot does not match last committed screenshot'
        if a is None:
            self.assertIsNone(b, msg)
        else:
            self.assertIsNotNone(b, msg)

        a_data = a.get_image_data()
        b_data = b.get_image_data()

        self.assertEqual(a_data.width, b_data.width, msg)
        self.assertEqual(a_data.height, b_data.height, msg)
        self.assertEqual(a_data.format, b_data.format, msg)
        self.assertEqual(a_data.pitch, b_data.pitch, msg)
        self.assert_buffer_equal(a_data.data, b_data.data, tolerance, msg)

    def assert_buffer_equal(self, a, b, tolerance=0, msg=None):
        if tolerance == 0:
            self.assertEqual(a, b, msg)

        self.assertEqual(len(a), len(b), msg)

        a = array.array('B', a)
        b = array.array('B', b)
        for (aa, bb) in zip(a, b):
            self.assertTrue(abs(aa - bb) <= tolerance, msg)

    def _take_screenshot(self):
        """
        Take a screenshot to allow visual verification.
        """
        screenshot_name = self._get_next_screenshot_name()
        screenshot_file_name = self._get_screenshot_session_file_name(screenshot_name)

        get_buffer_manager().get_color_buffer().get_image_data().save(screenshot_file_name)

        self._screenshots.append(screenshot_name)

    def _commit_screenshots(self):
        """
        Store the screenshots for reference if the test case is successful.
        """
        for screenshot_name in self._screenshots:
            shutil.copyfile(self._get_screenshot_session_file_name(screenshot_name),
                            self._get_screenshot_committed_file_name(screenshot_name))

    def _validate_screenshots(self):
        """
        Check the screenshot against regression reference images if available.
        """
        for screenshot_name in self._screenshots:
            committed_image = pyglet.image.load(self._get_screenshot_committed_file_name(screenshot_name))
            session_image = pyglet.image.load(self._get_screenshot_session_file_name(screenshot_name))
            self.assert_image_equal(committed_image, session_image)

    def _has_reference_screenshots(self):
        """
        Check whether there are screenshots from a previous successful run available to validate
        against. Use after taking all required screenshots. Also validates the number of required
        screenshots.
        """
        for screenshot_name in self._screenshots:
            if not os.path.isfile(self._get_screenshot_committed_file_name(screenshot_name)):
                return False
        else:
            return True

    def _get_next_screenshot_name(self):
        """
        Get the unique name for the next screenshot.
        """
        return '{}.{}.{}.{:03d}.png'.format(self.__class__.__module__,
                                        self.__class__.__name__,
                                        self._testMethodName,
                                        len(self._screenshots)+1)

    def _get_screenshot_session_file_name(self, screenshot_name):
        return os.path.join(session_screenshot_path, screenshot_name)

    def _get_screenshot_committed_file_name(self, screenshot_name):
        return os.path.join(committed_screenshot_path, screenshot_name)

def _ask_user_to_verify(description):
    failure_description = None
    print()
    print(description)
    while True:
        response = input('Passed [Yn]: ')
        if not response:
            break
        elif response in 'Nn':
            failure_description = input('Enter failure description: ')
            if not failure_description:
                failure_description = 'No description entered'
            break
        elif response in 'Yy':
            break
        else:
            print('Invalid response')
    return failure_description

@pytest.fixture
def interactive(request):
    """Fixture for interactive test cases. Returns an object that can be used for
    requesting interactive prompts and verifying screenshots.
    """
    return InteractiveFixture(request)



"""
Usage

    timeline.py sample [output_format]

Renders the media player's debug info to a format more suitable to postprocess
in a spreadsheets or other software, particularly to get a data visualization.

See output details in the manual.

Arguments
    sample: sample to report
    output_format : one of { "csv", "pkl"}, by default saves as .pkl (pickle)

The output will be written to session's output dir under
    reports/sample.timeline.[.pkl or .csv]

Example

    timeline.py small.mp4

will write the output to report/small.mp4.timeline.pkl

NOTE: .csv sample is currently not implemented
"""

import sys

from pyglet.media import instrumentation as ins
import fs
import mpexceptions


def usage():
    print(__doc__)
    sys.exit(1)


def main(sample, output_format):
    try:
        pathserv = fs.get_path_info_for_active_session()
    except mpexceptions.ExceptionUndefinedSamplesDir:
        print("The env var 'pyglet_mp_samples_dir' is not defined.")
        return 1
    except mpexceptions.ExceptionNoSessionIsActive:
        print("*** Error, no session active.")
        return 1

    try:
        save_timeline(pathserv, sample, output_format)
    except mpexceptions.ExceptionUnknownOutputFormat as ex:
        print("*** Error, unknown output_format: %s" % output_format)
    return 0


def save_timeline(pathserv, sample, output_format):
    dbg_file = pathserv.dbg_filename(sample)
    recorded_events = fs.pickle_load(dbg_file)
    tm = ins.TimelineBuilder(recorded_events, events_definition=ins.mp_events)
    timeline = tm.get_timeline()
    v = ins.timeline_postprocessing(timeline)
    if output_format == ".pkl":
        outfile = pathserv.report_filename(sample, "timeline.pkl", False)
        fs.pickle_save(v, outfile)
    elif output_format == ".csv":
        outfile = pathserv.report_filename(sample, "timeline.csv", False)
        #? todo: investigate packing multiple tables
        raise NotImplemented
    else:
        raise mpexceptions.ExceptionUnknownOutputFormat(output_format)


def sysargs_to_mainargs():
    """builds main args from sys.argv"""
    if len(sys.argv) < 2:
        usage()
    if sys.argv[1].startswith("--"):
        a = sys.argv.pop(1)
        if a.startswith("--help"):
            usage()
        else:
            print("Error unknown option:", a)
            usage()

    if len(sys.argv) < 2:
        print("*** Error, missing argument.\n")
        usage()
    sample = sys.argv.pop(1)

    output_format = ".pkl"
    if len(sys.argv) > 1:
        output_format = sys.argv.pop(1)

    return sample, output_format

if __name__ == '__main__':
    sample, output_format = sysargs_to_mainargs()
    main(sample, output_format)



from __future__ import annotations

import ctypes
from typing import Any, Callable, NoReturn, Sequence

import pyglet

__all__ = ['link_GL', 'link_AGL', 'link_GLX', 'link_WGL',
           'GLException', 'missing_function', 'decorate_function']

_debug_gl = pyglet.options['debug_gl']
_debug_gl_trace = pyglet.options['debug_gl_trace']
_debug_gl_trace_args = pyglet.options['debug_gl_trace_args']


class MissingFunctionException(Exception):  # noqa: N818
    def __init__(self, name: str, requires: str | None = None, suggestions: Sequence[str] | None=None) -> None:
        msg = f'{name} is not exported by the available OpenGL driver.'
        if requires:
            msg += f'  {requires} is required for this functionality.'
        if suggestions:
            msg += '  Consider alternative(s) {}.'.format(', '.join(suggestions))
        Exception.__init__(self, msg)


def missing_function(name: str, requires: str | None =None, suggestions: Sequence[str] | None=None) -> Callable:  # noqa: D103
    def MissingFunction(*_args, **_kwargs) -> NoReturn:  # noqa: ANN002, ANN003, N802
        raise MissingFunctionException(name, requires, suggestions)

    return MissingFunction


_int_types = (ctypes.c_int16, ctypes.c_int32)
if hasattr(ctypes, 'c_int64'):
    # Some builds of ctypes apparently do not have c_int64
    # defined; it's a pretty good bet that these builds do not
    # have 64-bit pointers.
    _int_types += (ctypes.c_int64,)
for t in _int_types:
    if ctypes.sizeof(t) == ctypes.sizeof(ctypes.c_size_t):
        c_ptrdiff_t = t


class c_void(ctypes.Structure):
    # c_void_p is a buggy return type, converting to int, so
    # POINTER(None) == c_void_p is actually written as
    # POINTER(c_void), so it can be treated as a real pointer.
    _fields_ = [('dummy', ctypes.c_int)]


class GLException(Exception):
    pass


def errcheck(result: Any, func: Callable, arguments: Sequence) -> Any:
    if _debug_gl_trace:
        try:
            name = func.__name__
        except AttributeError:
            name = repr(func)
        if _debug_gl_trace_args:
            trace_args = ', '.join([repr(arg)[:20] for arg in arguments])
            print(f'{name}({trace_args})')
        else:
            print(name)

    from pyglet import gl
    if not gl.current_context:
        raise GLException('No GL context; create a Window first')
    error = gl.glGetError()
    if error:
        # These are the 6 possible error codes we can get in opengl core 3.3+
        error_types = {
            gl.GL_INVALID_ENUM: "Invalid enum. An unacceptable value is specified for an enumerated argument.",
            gl.GL_INVALID_VALUE: "Invalid value. A numeric argument is out of range.",
            gl.GL_INVALID_OPERATION: "Invalid operation. The specified operation is not allowed in the current state.",
            gl.GL_INVALID_FRAMEBUFFER_OPERATION: "Invalid framebuffer operation. The framebuffer object is not "
                                                 "complete.",
            gl.GL_OUT_OF_MEMORY: "Out of memory. There is not enough memory left to execute the command.",
        }
        error_msg = error_types.get(error, "Unknown error")
        msg = f'(0x{error}): {error_msg}'
        raise GLException(msg)
    return result


def decorate_function(func: Callable, name: str) -> None:  # noqa: D103
    if _debug_gl and name not in ('glGetError',) and name[:3] not in ('glX', 'agl', 'wgl'):
        func.errcheck = errcheck
        func.__name__ = name


link_AGL = None
link_GLX = None
link_WGL = None

if pyglet.compat_platform in ('win32', 'cygwin'):
    from pyglet.gl.lib_wgl import link_GL, link_WGL
elif pyglet.compat_platform == 'darwin':
    from pyglet.gl.lib_agl import link_AGL, link_GL
else:
    from pyglet.gl.lib_glx import link_GL, link_GLX



from __future__ import annotations

import warnings
from ctypes import POINTER, byref, c_ubyte, cast, memmove
from typing import NamedTuple, Sequence

import pyglet
from pyglet import image
from pyglet.font import base
from pyglet.font.base import GlyphPosition
from pyglet.font.fontconfig import get_fontconfig
from pyglet.font.freetype_lib import (
    FT_LOAD_RENDER,
    FT_PIXEL_MODE_GRAY,
    FT_PIXEL_MODE_MONO,
    FT_STYLE_FLAG_BOLD,
    FT_STYLE_FLAG_ITALIC,
    FT_Byte,
    FT_Done_Face,
    FT_Face,
    FT_GlyphSlot,
    FT_Load_Glyph,
    FT_New_Face,
    FT_New_Memory_Face,
    FT_Reference_Face,
    FT_Set_Char_Size,
    f26p6_to_float,
    float_to_f26p6,
    ft_get_library, FT_LOAD_TARGET_MONO, FT_LOAD_COLOR, FT_FACE_FLAG_SCALABLE, FT_FACE_FLAG_FIXED_SIZES,
    FT_FACE_FLAG_COLOR, FT_Select_Size, FT_PIXEL_MODE_BGRA, FT_Get_Char_Index, FT_FACE_FLAG_KERNING, FT_Vector,
    FT_Get_Kerning, FT_KERNING_DEFAULT, FT_LOAD_NO_BITMAP, FT_Load_Char
)
from pyglet.font.harfbuzz import harfbuzz_available, get_resource_from_ft_font, get_harfbuzz_shaped_glyphs
from pyglet.util import asbytes, asstr


class FreeTypeGlyphRenderer(base.GlyphRenderer):
    def __init__(self, font: FreeTypeFont) -> None:
        super().__init__(font)
        self.font = font

        self._glyph_slot = None
        self._bitmap = None

        self._width = None
        self._height = None
        self._mode = None
        self._pitch = None

        self._baseline = None
        self._lsb = None
        self._advance_x = None

        self._data = None

    def _get_glyph_by_char(self, character: str) -> None:
        assert self.font
        assert len(character) == 1

        self._glyph_slot = self.font.get_glyph_slot(character)
        self._bitmap = self._glyph_slot.bitmap

    def _get_glyph_by_index(self, glyph_index: int) -> None:
        self._glyph_slot = self.font.get_glyph_slot_index(glyph_index)
        self._bitmap = self._glyph_slot.bitmap

    def _get_glyph_metrics(self) -> None:
        self._width = self._glyph_slot.bitmap.width
        self._height = self._glyph_slot.bitmap.rows
        self._mode = self._glyph_slot.bitmap.pixel_mode
        self._pitch = self._glyph_slot.bitmap.pitch

        self._baseline = self._height - self._glyph_slot.bitmap_top
        self._lsb = self._glyph_slot.bitmap_left
        self._advance_x = int(f26p6_to_float(self._glyph_slot.advance.x))

    def _expand_to_rgba(self, data: bytes, src_format: str, dst_format: str) -> bytes:
        """Expands data type to 4 components with putting values into A.

        Will re-evaluate on a better system later.
        """
        src_len = len(src_format)
        dst_len = len(dst_format)

        if src_len >= dst_len:
            return data

        expanded_data = bytearray(len(data) // src_len * dst_len)
        mapping = {c: i for i, c in enumerate(src_format)}

        for i in range(len(data) // src_len):
            default_value = data[i * src_len + 0] if src_len > 0 else 0

            for j, c in enumerate(dst_format):
                if c in mapping:
                    expanded_data[i * dst_len + j] = 255
                elif c == 'A':
                    # Default alpha to fully opaque
                    expanded_data[i * dst_len + j] = default_value
                else:
                    expanded_data[i * dst_len + j] = 255

        return bytes(expanded_data)

    def _get_bitmap_data(self) -> None:
        if self._mode == FT_PIXEL_MODE_MONO:
            # BCF fonts always render to 1 bit mono, regardless of render
            # flags. (freetype 2.3.5)
            self._convert_mono_to_gray_bitmap()
        elif self._mode == FT_PIXEL_MODE_GRAY:
            # Usual case
            assert self._glyph_slot.bitmap.num_grays == 256
            self._data = self._glyph_slot.bitmap.buffer
        elif self._mode == FT_PIXEL_MODE_BGRA:
            # as of freetype 2.5
            self._data = self._glyph_slot.bitmap.buffer
        else:
            msg = "Unsupported render mode for this glyph"
            raise base.FontException(msg)

    def _convert_mono_to_gray_bitmap(self) -> None:
        if self._bitmap.buffer:
            bitmap_data = cast(self._bitmap.buffer,
                               POINTER(c_ubyte * (self._pitch * self._height))).contents

            data = (c_ubyte * (self._width * self._height))()

            # Tightly pack the data, as freetype pads it.
            for y in range(self._height):
                for x in range(self._width):
                    byte = bitmap_data[y * self._pitch + (x // 8)]
                    bit = 7 - (x % 8)  # Data is MSB; left-most pixel in a byte has value 128.
                    data[y * self._width + x] = 255 if (byte & (1 << bit)) else 0
        else:
            # No pointer in the buffer, no default or fallback in this font.
            data = (c_ubyte * 0)()
        self._data = data
        self._pitch = self._width

    def _create_glyph(self) -> base.Glyph:
        # Textures should be a minimum of 1x1.
        if self._width == 0 and self._height == 0:
            width = 1
            height = 1
            data = bytes(bytearray([0, 0, 0, 0]))
            pitch = 4
        else:
            width = self._width
            height = self._height
            # If it's not in BGRA, convert it manually.
            if self._mode != FT_PIXEL_MODE_BGRA:
                size = self._width * self._height
                ptr = cast(self._data, POINTER(c_ubyte * size))
                data = self._expand_to_rgba(ptr.contents, 'R', 'BGRA')
                pitch = self._pitch * 4
            else:
                data = self._data
                pitch = self._pitch

        img = image.ImageData(width, height, "BGRA", data, pitch)

        glyph = self.font.create_glyph(img)
        glyph.set_bearings(self._baseline, self._lsb, self._advance_x)

        # In FT positive pitch means `down` flow, in Pyglet ImageData
        # negative values indicate a top-to-bottom arrangement. So pitch must be inverted.
        # Using negative pitch causes a CPU re-ordering. For now, swap texture coordinates for speed.
        if self._pitch > 0:
            t = list(glyph.tex_coords)
            glyph.tex_coords = t[9:12] + t[6:9] + t[3:6] + t[:3]

        return glyph

    def render_index(self, glyph_index: int):
        self._get_glyph_by_index(glyph_index)
        self._get_glyph_metrics()
        self._get_bitmap_data()
        return self._create_glyph()

    def render(self, text: str) -> base.Glyph:
        self._get_glyph_by_char(text[0])
        self._get_glyph_metrics()
        self._get_bitmap_data()
        return self._create_glyph()


class FreeTypeFontMetrics(NamedTuple):
    ascent: int
    descent: int


class MemoryFaceStore:
    _dict: dict[tuple[str, str, bool], FreeTypeMemoryFace]

    def __init__(self) -> None:
        self._dict = {}

    def add(self, face: FreeTypeMemoryFace) -> None:
        self._dict[face.name.lower(), face.weight, face.italic] = face

    def contains(self, name: str) -> bool:
        lname = name and name.lower() or ""
        return len([name for name, _, _ in self._dict if name == lname]) > 0

    def get(self, name: str, weight: str, italic: bool) -> FreeTypeMemoryFace | None:
        lname = name and name.lower() or ""
        return self._dict.get((lname, weight, italic), None)


class FreeTypeFont(base.Font):
    glyph_renderer_class = FreeTypeGlyphRenderer
    _glyph_renderer: FreeTypeGlyphRenderer

    # Map font (name, weight, italic) to FreeTypeMemoryFace
    _memory_faces = MemoryFaceStore()
    face: FreeTypeFace
    fallbacks: list[FreeTypeFont]

    def __init__(self, name: str, size: float, weight: str = "normal", italic: bool = False,
                 stretch: bool | str = False, dpi: int | None = None) -> None:
        super().__init__()
        self._name = name
        self.size = size
        self.weight = weight
        self.italic = italic
        self.stretch = stretch
        self.dpi = dpi or 96
        self.pixel_size = (self.size * self.dpi) // 72

        self._load_font_face()
        self.metrics = self.face.get_font_metrics(self.size, self.dpi)

        if pyglet.options.text_shaping == 'harfbuzz' and harfbuzz_available():
            self.hb_resource = get_resource_from_ft_font(self)

    @property
    def name(self) -> str:
        return self.face.family_name

    @property
    def ascent(self) -> int:
        return self.metrics.ascent

    @property
    def descent(self) -> int:
        return self.metrics.descent

    def add_fallback(self, font):
        self.fallbacks.append(font)

    def _get_slot_from_fallbacks(self, character: str) -> FT_GlyphSlot | None:
        """Checks all fallback fonts in order to find a valid glyph index."""
        # Check if fallback has this glyph, if so.
        for fallback_font in self.fallbacks:
            fb_index = fallback_font.face.get_character_index(character)
            if fb_index:
                fallback_font.face.set_char_size(self.size, self.dpi)
                return fallback_font.get_glyph_slot(character)

        return None

    def get_glyph_slot_index(self, glyph_index: int) -> FT_GlyphSlot:
        self.face.set_char_size(self.size, self.dpi)
        return self.face.get_glyph_slot(glyph_index)

    def get_glyph_slot(self, character: str) -> FT_GlyphSlot:
        glyph_index = self.face.get_character_index(character)
        # Glyph index does not exist, so check fallback fonts.
        if glyph_index == 0 and (self.name not in self.fallbacks):
            glyph_slot = self._get_slot_from_fallbacks(character)
            if glyph_slot is not None:
                return glyph_slot

        self.face.set_char_size(self.size, self.dpi)
        return self.face.get_glyph_slot(glyph_index)

    def _load_font_face(self) -> None:
        self.face = self._memory_faces.get(self._name, self.weight, self.italic)
        if self.face is None:
            self._load_font_face_from_system()

    def _load_font_face_from_system(self) -> None:
        match = get_fontconfig().find_font(self._name, self.size, self.weight, self.italic, self.stretch)
        if not match:
            msg = f"Could not match font '{self._name}'"
            raise base.FontException(msg)
        self.filename = match.file
        self.face = FreeTypeFace.from_fontconfig(match)

    @classmethod
    def have_font(cls: type[FreeTypeFont], name: str) -> bool:
        if cls._memory_faces.contains(name):
            return True

        return get_fontconfig().have_font(name)

    @classmethod
    def add_font_data(cls: type[FreeTypeFont], data: bytes) -> None:
        font_data = (FT_Byte * len(data))()
        memmove(font_data, data, len(data))

        face = FreeTypeMemoryFace(font_data, 0)
        cls._memory_faces.add(face)
        count = face.face_count
        # Some fonts may be a collection. Load each one.
        if count > 1:
            for i in range(1, count):
                face = FreeTypeMemoryFace(font_data, i)
                cls._memory_faces.add(face)

    def render_glyph_indices(self, indices: Sequence[int]):
        # Process any glyphs that have not been rendered.
        self._initialize_renderer()

        missing = set()
        for glyph_indice in set(indices):
            if glyph_indice not in self.glyphs:
                missing.add(glyph_indice)

        # Missing glyphs, get their info.
        for glyph_indice in missing:
            self.glyphs[glyph_indice] = self._glyph_renderer.render_index(glyph_indice)

    def get_glyphs(self, text: str) -> tuple[list[base.Glyph], list[base.GlyphPosition]]:
        """Create and return a list of Glyphs for `text`.

        If any characters do not have a known glyph representation in this
        font, a substitution will be made.

        Args:
            text:
                Text to render.
        """
        self._initialize_renderer()
        if pyglet.options.text_shaping == "harfbuzz" and harfbuzz_available():
            return get_harfbuzz_shaped_glyphs(self, text)
        else:
            glyphs = []  # glyphs that are committed.
            for idx, c in enumerate(text):
                # Get the glyph for 'c'.  Hide tabs (Windows and Linux render
                # boxes)
                if c == "\t":
                    c = " "  # noqa: PLW2901
                if c not in self.glyphs:
                    self.glyphs[c] = self._glyph_renderer.render(c)
                glyphs.append(self.glyphs[c])

            return glyphs, [GlyphPosition(0, 0, 0, 0)] * len(text)

    def get_text_size(self, text: str) -> tuple[int, int]:
        width = 0
        max_top = 0
        min_bottom = 0
        length = len(text)
        self.face.set_char_size(self.size, self.dpi)
        for i, char in enumerate(text):
            FT_Load_Char(self.face.ft_face, ord(char), FT_LOAD_NO_BITMAP)
            slot = self.face.ft_face.contents.glyph.contents

            if i == length-1:
                # Last glyph, use just the width.
                width += slot.metrics.width >> 6
            else:
                width += slot.advance.x >> 6

            glyph_top = slot.metrics.horiBearingY >> 6
            glyph_bottom = (slot.metrics.horiBearingY - slot.metrics.height) >> 6

            max_top = max(max_top, glyph_top)
            min_bottom = min(min_bottom, glyph_bottom)

        return width, max_top - min_bottom

class FreeTypeFace:
    """FreeType typographic face object.

    Keeps the reference count to the face at +1 as long as this object exists. If other objects
    want to keep a face without a reference to this object, they should increase the reference
    counter themselves and decrease it again when done.
    """
    _name: str
    ft_face: FT_Face

    def __init__(self, ft_face: FT_Face) -> None:
        assert ft_face is not None
        self.ft_face = ft_face
        self._get_best_name()

        self._italic = self.style_flags & FT_STYLE_FLAG_ITALIC != 0
        bold = self.style_flags & FT_STYLE_FLAG_BOLD != 0
        if bold:
            self._weight = "bold"
        else:
            # Sometimes it may have a weight, but FT_STYLE_FLAG_BOLD is not accurate. Check the font config.
            config = get_fontconfig()
            self._weight, italic, self._stretch = config.style_from_face(self.ft_face)
            if italic != self._italic:
                # Discrepancy in italics?
                self._italic = italic

    @classmethod
    def from_file(cls: type[FreeTypeFace], file_name: str) -> FreeTypeFace:
        ft_library = ft_get_library()
        ft_face = FT_Face()
        FT_New_Face(ft_library,
                    asbytes(file_name),
                    0,
                    byref(ft_face))
        return cls(ft_face)

    @classmethod
    def from_fontconfig(cls: type[FreeTypeFace], match):
        if match.face is not None:
            FT_Reference_Face(match.face)
            return cls(match.face)
        else:
            if not match.file:
                msg = f'No filename for "{match.name}"'
                raise base.FontException(msg)
            return cls.from_file(match.file)

    @property
    def name(self) -> str:
        return self._name

    @property
    def family_name(self) -> str:
        return asstr(self.ft_face.contents.family_name)

    @property
    def style_name(self) -> str:
        return asstr(self.ft_face.contents.style_name)

    @property
    def style_flags(self) -> int:
        return self.ft_face.contents.style_flags

    @property
    def bold(self) -> bool:
        return self.style_flags & FT_STYLE_FLAG_BOLD != 0

    @property
    def weight(self) -> str:
        return self._weight

    @property
    def italic(self) -> bool:
        return self._italic

    @property
    def face_flags(self) -> int:
        return self.ft_face.contents.face_flags

    def __del__(self) -> None:
        if self.ft_face is not None:
            FT_Done_Face(self.ft_face)
            self.ft_face = None

    def set_char_size(self, size: float, dpi: int) -> bool:
        if self.face_flags & FT_FACE_FLAG_SCALABLE:
            face_size = float_to_f26p6(size)
            FT_Set_Char_Size(self.ft_face, 0, face_size, dpi, dpi)

        elif self.face_flags & FT_FACE_FLAG_COLOR:
            if FT_Select_Size:
                FT_Select_Size(self.ft_face, 0)
            else:
                return False

        elif self.face_flags & FT_FACE_FLAG_FIXED_SIZES:
            if self.ft_face.contents.num_fixed_sizes:
                if FT_Select_Size:
                    FT_Select_Size(self.ft_face, 0)
                else:
                    return False
            else:
                warnings.warn(f"{self.name} no fixed sizes, but is flagged as a fixed size font.")

        return True

    def get_character_index(self, character: str) -> int:
        return FT_Get_Char_Index(self.ft_face, ord(character))

    def get_glyph_slot(self, glyph_index: int) -> FT_GlyphSlot:
        flags = FT_LOAD_RENDER
        if pyglet.options.text_antialiasing is False:
            flags |= FT_LOAD_TARGET_MONO

        if self.face_flags & FT_FACE_FLAG_COLOR:
            flags |= FT_LOAD_COLOR

        FT_Load_Glyph(self.ft_face, glyph_index, flags)
        return self.ft_face.contents.glyph.contents

    def get_font_metrics(self, size: float, dpi: int) -> FreeTypeFontMetrics:
        if self.set_char_size(size, dpi):
            metrics = self.ft_face.contents.size.contents.metrics
            if metrics.ascender == 0 and metrics.descender == 0:
                return self._get_font_metrics_workaround()

            return FreeTypeFontMetrics(ascent=int(f26p6_to_float(metrics.ascender)),
                                       descent=int(f26p6_to_float(metrics.descender)),
                                       )

        return self._get_font_metrics_workaround()

    def _get_font_metrics_workaround(self) -> FreeTypeFontMetrics:
        # Workaround broken fonts with no metrics.  Has been observed with
        # courR12-ISO8859-1.pcf.gz: "Courier" "Regular"
        #
        # None of the metrics fields are filled in, so render a glyph and
        # grab its height as the ascent, and make up an arbitrary
        # descent.
        i = self.get_character_index("X")
        self.get_glyph_slot(i)
        ascent = self.ft_face.contents.available_sizes.contents.height
        return FreeTypeFontMetrics(ascent=ascent,
                                   descent=-ascent // 4)  # arbitrary.

    def _get_best_name(self) -> None:
        self._name = asstr(self.ft_face.contents.family_name)
        self._get_font_family_from_ttf()

    def _get_font_family_from_ttf(self) -> None:
        # Replace Freetype's generic family name with TTF/OpenType specific
        # name if we can find one; there are some instances where Freetype
        # gets it wrong.

        return  # FIXME: This is broken

        if self.face_flags & FT_FACE_FLAG_SFNT:
            name = FT_SfntName()
            for i in range(FT_Get_Sfnt_Name_Count(self.ft_face)):
                try:
                    FT_Get_Sfnt_Name(self.ft_face, i, name)
                    if not (name.platform_id == TT_PLATFORM_MICROSOFT and
                            name.encoding_id == TT_MS_ID_UNICODE_CS):
                        continue
                    # name.string is not 0 terminated! use name.string_len
                    self._name = name.string.decode("utf-16be", "ignore")
                except:
                    continue


class FreeTypeMemoryFace(FreeTypeFace):
    def __init__(self, data: bytes, face_index: int = 0) -> None:
        self.font_data = data
        ft_library = ft_get_library()
        ft_face = FT_Face()
        FT_New_Memory_Face(ft_library,
                           self.font_data,
                           len(self.font_data),
                           face_index,
                           byref(ft_face))

        super().__init__(ft_face)

    @property
    def face_count(self) -> int:
        return self.ft_face.contents.num_faces



"""Base class for interactive tests spawning a basic window."""

from inspect import cleandoc

from pyglet import gl
from pyglet.window import Window

from tests.base.interactive import InteractiveTestCase


class WindowedTestCase(InteractiveTestCase):
    """
    Base class for tests that show a window, render something in that window and then ask a 
    question to the user whether the contents are correct. Also takes a screenshot when the
    test is passed, so it can run without interaction afterwards.
    """

    # Defaults
    window_size = 200, 200
    window_options = None
    window = None
    question = None
    take_screenshot = True

    # Methods to override in implementations
    def on_expose(self):
        pass

    def render(self):
        pass

    def draw(self):
        pass

    # Implementation of the base test class
    @classmethod
    def create_test_case(cls, name, description=None, decorators=None, **kwargs):
        def run_test(self):
            for name, value in kwargs.items():
                setattr(self, name, value)
            self._test_main()

        run_test.__name__ = name
        if description:
            run_test.__doc__ = cleandoc(description)
        if decorators:
            for decorator in decorators:
                run_test = decorator(run_test)
        setattr(cls, name, run_test)

    def _test_main(self):
        assert self.question

        self.window = w = Window(**self._get_window_options())
        try:
            w.push_handlers(self)
            self.render()
            w.set_visible()
            w.dispatch_events()

            self.user_verify(cleandoc(self.question), self.take_screenshot)

        finally:
            w.close()

    def _get_window_options(self):
        if self.window_options:
            options = self.window_options
        else:
            options = {}

        if 'width' not in options:
            options['width'] = self.window_size[0]
        if 'height' not in options:
            options['height'] = self.window_size[1]
        if 'visible' not in options:
            options['visible'] = False
        if 'resizable' not in options:
            options['resizable'] = True

        return options



#!/usr/bin/env python
import sys
import textwrap

import pyglet
import pyglet.app
import pyglet.display
import pyglet.window
from pyglet.gl import *
from pyglet.gl import gl_info

print('Pyglet:     %s' % pyglet.version)
print('Platform:   %s' % sys.platform)
print('Event loop: %s' % pyglet.app.PlatformEventLoop.__name__)

display = pyglet.display.get_display()
print('Display:    %s' % display.__class__.__name__)
print('Screens:')
for screen in display.get_screens():
    print('  %r' % screen)

print()
print('Creating default context...')
w = pyglet.window.Window(1, 1, visible=True)
print('Window:')
print('  %s' % w)

print('GL attributes:')
attrs = w.config.get_gl_attributes()
attrs = ' '.join(['%s=%s'%(name, value) for name, value in attrs])
print(' ', '\n  '.join(textwrap.wrap(attrs)))

print('GL version:', gl_info.get_version())
print('GL vendor:', gl_info.get_vendor())
print('GL renderer:', gl_info.get_renderer())
print('GL extensions:')
exts = ' '.join(gl_info.get_extensions())
print(' ', '\n  '.join(textwrap.wrap(exts)))

print()

context = w.context
print('Context is', context)

if "xlib" in globals() and isinstance(context, xlib.BaseXlibContext):
    from pyglet.gl import glx_info
    print('GLX %s direct'%(context.is_direct() and 'is' or 'is not'))
    if not glx_info.have_version(1, 1):
        print("GLX server version: 1.0")
    else:
        print('GLX server vendor:', glx_info.get_server_vendor())
        print('GLX server version:', glx_info.get_server_version())
        print('GLX server extensions:')
        exts = glx_info.get_server_extensions()
        print(' ', '\n  '.join(textwrap.wrap(' '.join(exts))))
        print('GLX client vendor:', glx_info.get_client_vendor())
        print('GLX client version:', glx_info.get_client_version())
        print('GLX client extensions:')
        exts = glx_info.get_client_extensions()
        print(' ', '\n  '.join(textwrap.wrap(' '.join(exts))))
        print('GLX extensions:')
        exts = glx_info.get_extensions()
        print(' ', '\n  '.join(textwrap.wrap(' '.join(exts))))
elif "win32" in globals() and isinstance(context, win32.Win32Context):
    from pyglet.gl import wgl_info
    if wgl_info.have_extension('WGL_EXT_extensions_string'):
        wgl_extensions = wgl_info.get_extensions()
        print('WGL extensions:')
        print('', '\n '.join(textwrap.wrap(' '.join(wgl_extensions))))
    else:
        print('WGL_EXT_extensions_string extension not available.')

w.close()



"""
Test the base class for interactive test cases.
"""

import glob
from tests import mock
import os
import pytest
import shutil
from ..base.interactive import InteractiveTestCase
import tempfile
import unittest

import pyglet
from pyglet import window
from pyglet.gl import *


@pytest.mark.requires_user_action
class InteractiveTestCaseTest(InteractiveTestCase):
    """
    Test the interactive test case base. Is an interactive test case itself, to be able to test it
    properly.
    """

    def setUp(self):
        self._patchers = []
        self._temporary_directories = []

    def tearDown(self):
        for patcher in self._patchers:
            patcher.stop()
        for directory in self._temporary_directories:
            shutil.rmtree(directory)

    def _patch_directory(self, target):
        directory = tempfile.mkdtemp()
        self._temporary_directories.append(directory)
        patcher = mock.patch(target, directory)
        self._patchers.append(patcher)
        patcher.start()
        return directory

    def _patch_screenshot_paths(self):
        self._session_screenshot_path = self._patch_directory('tests.base.interactive.session_screenshot_path')
        self._committed_screenshot_path = self._patch_directory('tests.base.interactive.committed_screenshot_path')

    def test_single_method(self):
        class _Test(InteractiveTestCase):
            test1_ran = False

            def test_1(self):
                _Test.test1_ran = True

        tests = unittest.defaultTestLoader.loadTestsFromTestCase(_Test)
        self.assertIsNotNone(tests)
        self.assertEqual(tests.countTestCases(), 1)

        result = unittest.TestResult()
        tests.run(result)

        self.assertTrue(_Test.test1_ran, 'Test should have run')

    def test_multiple_methods(self):
        class _Test(InteractiveTestCase):
            test1_ran = False
            test2_ran = False

            def test_1(self):
                _Test.test1_ran = True

            def test_2(self):
                _Test.test2_ran = True

        tests = unittest.defaultTestLoader.loadTestsFromTestCase(_Test)
        self.assertIsNotNone(tests)
        self.assertEqual(tests.countTestCases(), 2)

        result = unittest.TestResult()
        tests.run(result)

        self.assertTrue(_Test.test1_ran, 'Test 1 should have run')
        self.assertTrue(_Test.test2_ran, 'Test 2 should have run')

    def test_user_verify_passed(self):
        class _Test(InteractiveTestCase):
            test1_ran = False

            def test_1(self):
                _Test.test1_ran = True

                self.user_verify('Just press Enter', take_screenshot=False)

        tests = unittest.defaultTestLoader.loadTestsFromTestCase(_Test)
        self.assertIsNotNone(tests)
        self.assertEqual(tests.countTestCases(), 1)

        result = unittest.TestResult()
        tests.run(result)

        self.assertTrue(_Test.test1_ran, 'Test should have run')
        self.assertEqual(len(result.failures), 0, 'Not expecting failures')
        self.assertEqual(len(result.errors), 0, 'Not expecting errors')
        self.assertEqual(result.testsRun, 1, 'Expected 1 test run')

        self.user_verify('Did I ask you to press Enter?', take_screenshot=False)

    def test_user_verify_failed(self):
        class _Test(InteractiveTestCase):
            test1_ran = False

            def test_1(self):
                _Test.test1_ran = True

                self.user_verify('Enter "n" and then enter reason "abcd"', take_screenshot=False)

        tests = unittest.defaultTestLoader.loadTestsFromTestCase(_Test)
        self.assertIsNotNone(tests)
        self.assertEqual(tests.countTestCases(), 1)

        result = unittest.TestResult()
        tests.run(result)

        self.assertTrue(_Test.test1_ran, 'Test should have run')
        self.assertEqual(len(result.failures), 1, 'Expected 1 test failure')
        self.assertEqual(len(result.errors), 0, 'Not expecting errors')
        self.assertEqual(result.testsRun, 1, 'Expected 1 test run')

        self.assertIn('AssertionError: abcd', result.failures[0][1], 'Did not get failure message entered by user.')

    def test_verify_commits_screenshot_on_user_passed(self):
        class _Test(InteractiveTestCase):
            def test_1(self):
                w = window.Window(200, 200)
                w.switch_to()
                glClearColor(1, 0, 1, 1)
                glClear(GL_COLOR_BUFFER_BIT)
                w.flip()

                self.user_verify('Please choose yes (or press Enter)')
                w.close()

        self._patch_screenshot_paths()

        tests = unittest.defaultTestLoader.loadTestsFromTestCase(_Test)
        self.assertIsNotNone(tests)
        self.assertEqual(tests.countTestCases(), 1)

        result = unittest.TestResult()
        tests.run(result)

        self.assertEqual(len(result.failures), 0, 'Not expecting failures')
        self.assertEqual(len(result.errors), 0, 'Not expecting errors')
        self.assertEqual(result.testsRun, 1, 'Expected 1 test run')

        files = glob.glob(os.path.join(self._session_screenshot_path, '*.png'))
        self.assertEqual(len(files), 1, 'Screenshot not stored in session directory')
        self.assertIn('tests.interactive.test_interactive_test_base._Test.test_1.001.png', files[0])

        files = glob.glob(os.path.join(self._committed_screenshot_path, '*.png'))
        self.assertEqual(len(files), 1, 'Screenshot not committed')
        self.assertIn('tests.interactive.test_interactive_test_base._Test.test_1.001.png', files[0])

    @mock.patch('tests.interactive.interactive_test_base.interactive', False)
    def test_screenshot_taken_but_not_committed_on_noninteractive_failure(self):
        class _Test(InteractiveTestCase):
            def test_1(self):
                w = window.Window(200, 200)
                w.switch_to()
                glClearColor(1, 0, 1, 1)
                glClear(GL_COLOR_BUFFER_BIT)
                w.flip()

                self.user_verify('Empty window')
                w.close()

                self.fail('Test failed')

        self._patch_screenshot_paths()

        tests = unittest.defaultTestLoader.loadTestsFromTestCase(_Test)
        self.assertIsNotNone(tests)
        self.assertEqual(tests.countTestCases(), 1)

        result = unittest.TestResult()
        tests.run(result)

        self.assertEqual(len(result.failures), 1, 'Expecting 1 failure')
        self.assertEqual(len(result.errors), 0, 'Not expecting errors')
        self.assertEqual(result.testsRun, 1, 'Expected 1 test run')

        files = glob.glob(os.path.join(self._session_screenshot_path, '*.png'))
        self.assertEqual(len(files), 1, 'Screenshot not stored in session directory')
        self.assertIn('tests.interactive.test_interactive_test_base._Test.test_1.001.png', files[0])

        files = glob.glob(os.path.join(self._committed_screenshot_path, '*.png'))
        self.assertEqual(len(files), 0, 'Screenshot should not have been comitted')

    @mock.patch('tests.interactive.interactive_test_base.interactive', False)
    @mock.patch('tests.interactive.interactive_test_base.allow_missing_screenshots', True)
    def test_screenshot_taken_but_not_committed_on_noninteractive_pass(self):
        class _Test(InteractiveTestCase):
            def test_1(self):
                w = window.Window(200, 200)
                w.switch_to()
                glClearColor(1, 0, 1, 1)
                glClear(GL_COLOR_BUFFER_BIT)
                w.flip()

                self.user_verify('Empty window')
                w.close()

        self._patch_screenshot_paths()

        tests = unittest.defaultTestLoader.loadTestsFromTestCase(_Test)
        self.assertIsNotNone(tests)
        self.assertEqual(tests.countTestCases(), 1)

        result = unittest.TestResult()
        tests.run(result)

        self.assertEqual(len(result.failures), 0, 'Not expecting failures')
        self.assertEqual(len(result.errors), 0, 'Not expecting errors')
        self.assertEqual(result.testsRun, 1, 'Expected 1 test run')

        files = glob.glob(os.path.join(self._session_screenshot_path, '*.png'))
        self.assertEqual(len(files), 1, 'Screenshot not stored in session directory')
        self.assertIn('tests.interactive.test_interactive_test_base._Test.test_1.001.png', files[0])

        files = glob.glob(os.path.join(self._committed_screenshot_path, '*.png'))
        self.assertEqual(len(files), 0, 'Screenshot should not have been comitted')

    @mock.patch('tests.interactive.interactive_test_base.interactive', False)
    def test_fails_on_missing_screenshot_on_noninteractive_pass(self):
        class _Test(InteractiveTestCase):
            def test_1(self):
                w = window.Window(200, 200)
                w.switch_to()
                glClearColor(1, 0, 1, 1)
                glClear(GL_COLOR_BUFFER_BIT)
                w.flip()

                self.user_verify('Empty window')
                w.close()

        self._patch_screenshot_paths()

        tests = unittest.defaultTestLoader.loadTestsFromTestCase(_Test)
        self.assertIsNotNone(tests)
        self.assertEqual(tests.countTestCases(), 1)

        result = unittest.TestResult()
        tests.run(result)

        self.assertEqual(len(result.failures), 1, 'Expecting 1 failure')
        self.assertEqual(len(result.errors), 0, 'Not expecting errors')
        self.assertEqual(result.testsRun, 1, 'Expected 1 test run')

        files = glob.glob(os.path.join(self._session_screenshot_path, '*.png'))
        self.assertEqual(len(files), 1, 'Screenshot not stored in session directory')
        self.assertIn('tests.interactive.test_interactive_test_base._Test.test_1.001.png', files[0])

        files = glob.glob(os.path.join(self._committed_screenshot_path, '*.png'))
        self.assertEqual(len(files), 0, 'Screenshot should not have been comitted')


    def test_screenshot_taken_but_not_committed_on_user_failure(self):
        class _Test(InteractiveTestCase):
            def test_1(self):
                w = window.Window(200, 200)
                w.switch_to()
                glClearColor(1, 0, 1, 1)
                glClear(GL_COLOR_BUFFER_BIT)
                w.flip()

                try:
                    self.user_verify('Please select "n" and enter any reason')
                finally:
                    w.close()


        self._patch_screenshot_paths()

        tests = unittest.defaultTestLoader.loadTestsFromTestCase(_Test)
        self.assertIsNotNone(tests)
        self.assertEqual(tests.countTestCases(), 1)

        result = unittest.TestResult()
        tests.run(result)

        self.assertEqual(len(result.failures), 1, 'Expecting 1 failure')
        self.assertEqual(len(result.errors), 0, 'Not expecting errors')
        self.assertEqual(result.testsRun, 1, 'Expected 1 test run')

        files = glob.glob(os.path.join(self._session_screenshot_path, '*.png'))
        self.assertEqual(len(files), 1, 'Screenshot not stored in session directory')
        self.assertIn('tests.interactive.test_interactive_test_base._Test.test_1.001.png', files[0])

        files = glob.glob(os.path.join(self._committed_screenshot_path, '*.png'))
        self.assertEqual(len(files), 0, 'Screenshot should not have been committed')

    @mock.patch('tests.interactive.interactive_test_base.interactive', False)
    def test_screenshot_does_not_match(self):
        class _Test(InteractiveTestCase):
            def test_1(self):
                w = window.Window(200, 200)
                w.switch_to()
                glClearColor(0, 0, 1, 1)
                glClear(GL_COLOR_BUFFER_BIT)
                w.flip()

                self.user_verify('Empty window')
                w.close()

        self._patch_screenshot_paths()

        # Copy non matching screenshot
        screenshot_name = 'tests.interactive.test_interactive_test_base._Test.test_1.001.png'
        original_screenshot = os.path.join(os.path.dirname(__file__), '..', 'data', 'images', screenshot_name)
        committed_screenshot = os.path.join(self._committed_screenshot_path, screenshot_name)
        shutil.copy(original_screenshot, committed_screenshot)

        # Start the test
        tests = unittest.defaultTestLoader.loadTestsFromTestCase(_Test)
        self.assertIsNotNone(tests)
        self.assertEqual(tests.countTestCases(), 1)

        result = unittest.TestResult()
        tests.run(result)

        self.assertEqual(len(result.failures), 1, 'Expecting 1 failure')
        self.assertEqual(len(result.errors), 0, 'Not expecting errors')
        self.assertEqual(result.testsRun, 1, 'Expected 1 test run')

        files = glob.glob(os.path.join(self._session_screenshot_path, '*.png'))
        self.assertEqual(len(files), 1, 'Screenshot not stored in session directory')
        self.assertIn('tests.interactive.test_interactive_test_base._Test.test_1.001.png', files[0])

        # Verify committed image not changed
        original_image = pyglet.image.load(original_screenshot)
        committed_image = pyglet.image.load(committed_screenshot)
        self.assert_image_equal(original_image, committed_image, msg='Committed image should not be overwritten')

    @mock.patch('tests.interactive.interactive_test_base.interactive', False)
    def test_screenshot_matches(self):
        class _Test(InteractiveTestCase):
            def test_1(self):
                w = window.Window(200, 200)
                w.switch_to()
                glClearColor(1, 0, 1, 1)
                glClear(GL_COLOR_BUFFER_BIT)
                w.flip()

                self.user_verify('Empty window')
                w.close()

        self._patch_screenshot_paths()

        # Copy matching screenshot
        screenshot_name = 'tests.interactive.test_interactive_test_base._Test.test_1.001.png'
        original_screenshot = os.path.join(os.path.dirname(__file__), '..', 'data', 'images', screenshot_name)
        committed_screenshot = os.path.join(self._committed_screenshot_path, screenshot_name)
        shutil.copy(original_screenshot, committed_screenshot)

        # Start the test
        tests = unittest.defaultTestLoader.loadTestsFromTestCase(_Test)
        self.assertIsNotNone(tests)
        self.assertEqual(tests.countTestCases(), 1)

        result = unittest.TestResult()
        tests.run(result)

        self.assertEqual(len(result.failures), 0, 'Not expecting failures')
        self.assertEqual(len(result.errors), 0, 'Not expecting errors')
        self.assertEqual(result.testsRun, 1, 'Expected 1 test run')

        files = glob.glob(os.path.join(self._session_screenshot_path, '*.png'))
        self.assertEqual(len(files), 1, 'Screenshot not stored in session directory')
        self.assertIn('tests.interactive.test_interactive_test_base._Test.test_1.001.png', files[0])





from __future__ import annotations

import ctypes
from ctypes import byref, sizeof
from typing import TYPE_CHECKING

from pyglet.libs.win32 import _gdi32, _shcore, _user32
from pyglet.libs.win32.constants import (
    CDS_FULLSCREEN,
    DISP_CHANGE_SUCCESSFUL,
    DISPLAYCONFIG_DEVICE_INFO_GET_SOURCE_NAME,
    DISPLAYCONFIG_DEVICE_INFO_GET_TARGET_NAME,
    ENUM_CURRENT_SETTINGS,
    LOGPIXELSX,
    LOGPIXELSY,
    MONITORINFOF_PRIMARY,
    QDC_ONLY_ACTIVE_PATHS,
    USER_DEFAULT_SCREEN_DPI,
    WINDOWS_8_1_OR_GREATER,
    WINDOWS_10_CREATORS_UPDATE_OR_GREATER,
    WINDOWS_VISTA_OR_GREATER,
)
from pyglet.libs.win32.context_managers import device_context
from pyglet.libs.win32.types import (
    DEVMODE,
    DISPLAY_DEVICEW,
    DISPLAYCONFIG_PATH_INFO,
    DISPLAYCONFIG_SOURCE_DEVICE_NAME,
    DISPLAYCONFIG_TARGET_DEVICE_NAME,
    DPI_AWARENESS_CONTEXT_PER_MONITOR_AWARE_V2,
    MONITORENUMPROC,
    MONITORINFOEX,
    PROCESS_PER_MONITOR_DPI_AWARE,
    UINT,
    UINT32,
)

from .base import Canvas, Display, Screen, ScreenMode

if TYPE_CHECKING:
    from ctypes.wintypes import HDC, HMONITOR, LPARAM, LPRECT


def set_dpi_awareness() -> None:
    """Setting DPI varies per Windows version.

    .. note:: DPI awareness needs to be set before Window, Display, or Screens are initialized.
    """
    if WINDOWS_10_CREATORS_UPDATE_OR_GREATER:
        _user32.SetProcessDpiAwarenessContext(DPI_AWARENESS_CONTEXT_PER_MONITOR_AWARE_V2)
    elif WINDOWS_8_1_OR_GREATER:  # 8.1 and above allows per monitor DPI.
        _shcore.SetProcessDpiAwareness(PROCESS_PER_MONITOR_DPI_AWARE)
    elif WINDOWS_VISTA_OR_GREATER:  # Only System wide DPI
        _user32.SetProcessDPIAware()


set_dpi_awareness()


class Win32Display(Display):  # noqa: D101
    def get_default_screen(self) -> Screen:
        screens = self.get_screens()
        for screen in screens:
            if screen.is_primary:
                return screen

        return screens[0]

    def get_screens(self) -> list[Win32Screen]:
        screens = []

        def enum_proc(hMonitor: HMONITOR, hdcMonitor: HDC, lprcMonitor: LPRECT, dwData: LPARAM) -> bool:  # noqa: N803, ARG001
            r = lprcMonitor.contents
            width = r.right - r.left
            height = r.bottom - r.top
            screens.append(
                Win32Screen(self, hMonitor, r.left, r.top, width, height))
            return True

        enum_proc_ptr = MONITORENUMPROC(enum_proc)
        _user32.EnumDisplayMonitors(None, None, enum_proc_ptr, 0)
        return screens


class Win32Screen(Screen):  # noqa: D101
    _handle: HMONITOR
    _initial_mode = None

    def __init__(self, display: Win32Display, handle: HMONITOR, x: int, y: int, width: int, height: int) -> None:  # noqa: D107
        super().__init__(display, x, y, width, height)
        self._handle = handle
        self._device_name = self.get_device_name()  # \\.\DISPLAY1
        self._friendly_name = self._get_friendly_name()

    @property
    def is_primary(self) -> bool:
        """If the screen is considered the primary according to the operating system."""
        info = self._get_monitor_info()
        return info.dwFlags & MONITORINFOF_PRIMARY

    def _get_friendly_name_display_config_api(self) -> str:
        """Get the friendly name of a monitor using the newer Display Configuration API.

        This API is meant to replace EnumDisplayDevicesW, and should be more accurate.

        Requires Windows Vista or higher.
        """
        path_count = UINT32()
        mode_count = UINT32()

        result = _user32.GetDisplayConfigBufferSizes(
            QDC_ONLY_ACTIVE_PATHS, ctypes.byref(path_count), ctypes.byref(mode_count),
        )
        if result != 0:
            return "Unknown"

        paths = (DISPLAYCONFIG_PATH_INFO * path_count.value)()
        modes = ctypes.create_string_buffer(64 * mode_count.value)  # dummy buffer

        result = _user32.QueryDisplayConfig(
            QDC_ONLY_ACTIVE_PATHS, ctypes.byref(path_count), paths, ctypes.byref(mode_count), modes, 0,
        )
        if result != 0:
            return "Unknown"

        for i in range(path_count.value):
            path = paths[i]

            source_name = DISPLAYCONFIG_SOURCE_DEVICE_NAME()
            source_name.header.adapterId = path.sourceInfo.adapterId
            source_name.header.id = path.sourceInfo.id
            source_name.header.type = DISPLAYCONFIG_DEVICE_INFO_GET_SOURCE_NAME
            source_name.header.size = ctypes.sizeof(source_name)

            result = _user32.DisplayConfigGetDeviceInfo(ctypes.byref(source_name.header))
            if result != 0:
                continue

            if source_name.viewGdiDeviceName != self._device_name:
                continue

            if not path.targetInfo.targetAvailable:
                  continue

            target_name = DISPLAYCONFIG_TARGET_DEVICE_NAME()

            target_name.header.adapterId = path.targetInfo.adapterId
            target_name.header.id = path.targetInfo.id
            target_name.header.type = DISPLAYCONFIG_DEVICE_INFO_GET_TARGET_NAME
            target_name.header.size = ctypes.sizeof(target_name)

            if _user32.DisplayConfigGetDeviceInfo(ctypes.byref(target_name.header)) == 0:
                return target_name.monitorFriendlyDeviceName

        return "Unknown"

    def _get_friendly_name(self) -> str:
        if WINDOWS_VISTA_OR_GREATER:
            return self._get_friendly_name_display_config_api()

        dd = DISPLAY_DEVICEW()
        dd.cb = ctypes.sizeof(dd)
        if _user32.EnumDisplayDevicesW(self._device_name, 0, ctypes.byref(dd), 0):
            return dd.DeviceString

        return "Unknown"

    def get_matching_configs(self, template):
        with device_context(None) as hdc:
            canvas = Win32Canvas(self.display, 0, hdc)
            configs = template.match(canvas)
            # XXX deprecate config's being screen-specific
            for config in configs:
                config.screen = self

        return configs

    def _get_monitor_info(self) -> MONITORINFOEX:
        info = MONITORINFOEX()
        info.cbSize = sizeof(MONITORINFOEX)
        _user32.GetMonitorInfoW(self._handle, byref(info))
        return info

    def get_display_id(self) -> str:
        return self._device_name

    def get_monitor_name(self) -> str:
        return self._friendly_name

    def get_device_name(self) -> str:
        info = self._get_monitor_info()
        return info.szDevice

    def get_dpi(self) -> int:
        if WINDOWS_8_1_OR_GREATER:
            xdpi = UINT()
            ydpi = UINT()
            _shcore.GetDpiForMonitor(self._handle, 0, byref(xdpi), byref(ydpi))
            xdpi, ydpi = xdpi.value, ydpi.value
        else:
            dc = _user32.GetDC(None)
            xdpi = _gdi32.GetDeviceCaps(dc, LOGPIXELSX)
            ydpi = _gdi32.GetDeviceCaps(dc, LOGPIXELSY)
            _user32.ReleaseDC(0, dc)

        return xdpi

    def get_scale(self) -> float:
        xdpi = self.get_dpi()
        return xdpi / USER_DEFAULT_SCREEN_DPI

    def get_modes(self) -> list[Win32ScreenMode]:
        device_name = self.get_device_name()
        i = 0
        modes = []
        while True:
            mode = DEVMODE()
            mode.dmSize = sizeof(DEVMODE)
            r = _user32.EnumDisplaySettingsW(device_name, i, byref(mode))
            if not r:
                break

            modes.append(Win32ScreenMode(self, mode))
            i += 1

        return modes

    def get_mode(self) -> Win32ScreenMode:
        mode = DEVMODE()
        mode.dmSize = sizeof(DEVMODE)
        _user32.EnumDisplaySettingsW(self.get_device_name(),
                                     ENUM_CURRENT_SETTINGS,
                                     byref(mode))
        return Win32ScreenMode(self, mode)

    def set_mode(self, mode: Win32ScreenMode) -> None:
        assert mode.screen is self

        if not self._initial_mode:
            self._initial_mode = self.get_mode()
        r = _user32.ChangeDisplaySettingsExW(self.get_device_name(),
                                             byref(mode._mode),
                                             None,
                                             CDS_FULLSCREEN,
                                             None)
        if r == DISP_CHANGE_SUCCESSFUL:
            self.width = mode.width
            self.height = mode.height

    def restore_mode(self) -> None:
        if self._initial_mode:
            self.set_mode(self._initial_mode)


_win32_scale_name = {
    0: "default",
    1: "center",
    2: "stretch",
}
class Win32ScreenMode(ScreenMode):  # noqa: D101
    def __init__(self, screen: Win32Screen, mode: DEVMODE) -> None:  # noqa: D107
        super().__init__(screen)
        self._mode = mode
        self.width = mode.dmPelsWidth
        self.height = mode.dmPelsHeight
        self.depth = mode.dmBitsPerPel
        self.rate = mode.dmDisplayFrequency
        self.scaling = mode.dmDisplayFixedOutput

    def __repr__(self) -> str:
        return (f'{self.__class__.__name__}(width={self.width!r}, height={self.height!r}, depth={self.depth!r},'
                f'rate={self.rate}, scaling={_win32_scale_name.get(self.scaling)})')

class Win32Canvas(Canvas):  # noqa: D101
    def __init__(self, display: Win32Display, hwnd: HWND, hdc: HDC) -> None:  # noqa: D107
        super().__init__(display)
        self.hwnd = hwnd
        self.hdc = hdc



"""Functions for loading dynamic libraries.

These extend and correct ctypes functions.
"""
from __future__ import annotations

import os
import re
import sys
import contextlib

import ctypes
import ctypes.util

import pyglet
from typing import NoReturn, Callable, Any

_debug_lib = pyglet.options['debug_lib']
_debug_trace = pyglet.options['debug_trace']

_is_pyglet_doc_run = getattr(sys, "is_pyglet_doc_run", False)

if pyglet.options['search_local_libs']:
    script_path = pyglet.resource.get_script_home()
    cwd = os.getcwd()
    _local_lib_paths = [script_path, os.path.join(script_path, 'lib'), os.path.join(cwd, 'lib')]
    if pyglet.compat_platform == 'win32':
        os.environ["PATH"] += os.pathsep + os.pathsep.join(_local_lib_paths)
else:
    _local_lib_paths = None


class _TraceFunction:
    def __init__(self, func: Callable) -> None:
        self.__dict__['_func'] = func

    def __str__(self) -> str:
        return self._func.__name__

    def __call__(self, *args, **kwargs):
        return self._func(*args, **kwargs)

    def __getattr__(self, name: str) -> Any:
        return getattr(self._func, name)

    def __setattr__(self, name: str, value: object) -> None:
        setattr(self._func, name, value)


class _TraceLibrary:
    def __init__(self, library: Any) -> None:
        self._library = library
        print(library)

    def __getattr__(self, name: str) -> Callable:
        func = getattr(self._library, name)
        return _TraceFunction(func)


if _is_pyglet_doc_run:
    class LibraryMock:
        """Mock library used when generating documentation."""
        def __getattr__(self, name: str):
            return LibraryMock()

        def __setattr__(self, name: str, value) -> None:
            pass

        def __call__(self, *args, **kwargs):
            return LibraryMock()

        def __rshift__(self, other: LibraryMock):
            return 0


class LibraryLoader:  # noqa: D101

    platform = pyglet.compat_platform
    # this is only for library loading, don't include it in pyglet.platform
    if platform == 'cygwin':
        platform = 'win32'

    def load_library(self, *names: str, **kwargs):
        """Find and load a library.

        More than one name can be specified, they will be tried in order.
        Platform-specific library names (given as kwargs) are tried first.

        Raises ImportError if library is not found.
        """
        if _is_pyglet_doc_run:
            return LibraryMock()

        if 'framework' in kwargs and self.platform == 'darwin':
            return self.load_framework(kwargs['framework'])

        if not names:
            msg = "No library name specified"
            raise ImportError(msg)

        platform_names = kwargs.get(self.platform, [])
        if isinstance(platform_names, str):
            platform_names = [platform_names]
        elif type(platform_names) is tuple:
            platform_names = list(platform_names)

        if self.platform.startswith('linux'):
            for name in names:
                libname = self.find_library(name)
                platform_names.append(libname or f'lib{name}.so')

        platform_names.extend(names)
        for name in platform_names:
            try:
                lib = ctypes.cdll.LoadLibrary(name)
                if _debug_lib:
                    print(name, self.find_library(name))
                if _debug_trace:
                    lib = _TraceLibrary(lib)
                return lib
            except OSError as o: # noqa: PERF203
                path = self.find_library(name)
                if path:
                    try:
                        lib = ctypes.cdll.LoadLibrary(path)
                        if _debug_lib:
                            print(path)
                        if _debug_trace:
                            lib = _TraceLibrary(lib)
                        return lib
                    except OSError as e:
                        if _debug_lib:
                            print(f"Unexpected error loading library {name}: {e!s}")
                elif self.platform == "win32" and o.winerror != 126 and _debug_lib:
                    print(f"Unexpected error loading library {name}: {o!s}")

        msg = f'Library "{names[0]}" not found.'
        raise ImportError(msg)

    def find_library(self, name: str) -> str | None:
        return ctypes.util.find_library(name)

    @staticmethod
    def load_framework(_name: str) -> NoReturn:
        msg = "Can't load framework on this platform."
        raise RuntimeError(msg)


class MacOSLibraryLoader(LibraryLoader):  # noqa: D101
    def __init__(self) -> None:  # noqa: D107
        if 'LD_LIBRARY_PATH' in os.environ:
            self.ld_library_path = os.environ['LD_LIBRARY_PATH'].split(':')
        else:
            self.ld_library_path = []

        if _local_lib_paths:
            # search first for local libs
            self.ld_library_path = _local_lib_paths + self.ld_library_path
            os.environ['LD_LIBRARY_PATH'] = ':'.join(self.ld_library_path)

        if 'DYLD_LIBRARY_PATH' in os.environ:
            self.dyld_library_path = os.environ['DYLD_LIBRARY_PATH'].split(':')
        else:
            self.dyld_library_path = []

        if 'DYLD_FALLBACK_LIBRARY_PATH' in os.environ:
            self.dyld_fallback_library_path = os.environ['DYLD_FALLBACK_LIBRARY_PATH'].split(':')
        else:
            self.dyld_fallback_library_path = [os.path.expanduser('~/lib'), '/usr/local/lib', '/usr/lib']

            # Homebrew path on Apple Silicon is no longer in local.
            if 'HOMEBREW_PREFIX' in os.environ:
                # if HOMEBREW_PREFIX is defined, add its lib directory.
                brew_lib_path = os.path.join(os.environ['HOMEBREW_PREFIX'], 'lib')
                if os.path.exists(brew_lib_path):
                    self.dyld_fallback_library_path.append(brew_lib_path)
            else:
                # Check the typical path if the environmental variable is missing.
                if os.path.exists('/opt/homebrew/lib'):
                    self.dyld_fallback_library_path.append('/opt/homebrew/lib')


    def find_library(self, path: str) -> str | None:
        """Implements the dylib search as specified in Apple documentation:

        http://developer.apple.com/library/content/documentation/DeveloperTools/Conceptual/DynamicLibraries/100-Articles/DynamicLibraryUsageGuidelines.html

        Before commencing the standard search, the method first checks
        the bundle's ``Frameworks`` directory if the application is running
        within a bundle (OS X .app).
        """  # noqa: D415
        libname = os.path.basename(path)
        search_path = []

        if '.dylib' not in libname:
            libname = 'lib' + libname + '.dylib'

        # py2app support
        if getattr(sys, 'frozen', None) == 'macosx_app' and 'RESOURCEPATH' in os.environ:
            search_path.append(os.path.join(os.environ['RESOURCEPATH'],
                                            '..',
                                            'Frameworks',
                                            libname))

        # conda support
        if os.environ.get('CONDA_PREFIX', False):
            search_path.append(os.path.join(os.environ['CONDA_PREFIX'], 'lib', libname))

        # pyinstaller.py sets sys.frozen to True, and puts dylibs in
        # Contents/macOS, which path pyinstaller puts in sys._MEIPASS
        if getattr(sys, 'frozen', False) and (meipass := getattr(sys, '_MEIPASS', None)):
            search_path.append(os.path.join(meipass, libname))

        # conda support
        if os.environ.get('CONDA_PREFIX', False):
            search_path.append(os.path.join(os.environ['CONDA_PREFIX'], 'lib', libname))

        if '/' in path:
            search_path.extend([os.path.join(p, libname) for p in self.dyld_library_path])
            search_path.append(path)
            search_path.extend([os.path.join(p, libname) for p in self.dyld_fallback_library_path])
        else:
            search_path.extend([os.path.join(p, libname) for p in self.ld_library_path])
            search_path.extend([os.path.join(p, libname) for p in self.dyld_library_path])
            search_path.append(path)
            search_path.extend([os.path.join(p, libname) for p in self.dyld_fallback_library_path])

        for path in search_path:
            if os.path.exists(path):
                return path

        return None

    @staticmethod
    def load_framework(name: str) -> ctypes.CDLL | _TraceLibrary:
        path = ctypes.util.find_library(name)

        # Hack for compatibility with macOS > 11.0  # noqa: FIX004
        if path is None:
            frameworks = {
                'AGL': '/System/Library/Frameworks/AGL.framework/AGL',
                'IOKit': '/System/Library/Frameworks/IOKit.framework/IOKit',
                'OpenAL': '/System/Library/Frameworks/OpenAL.framework/OpenAL',
                'OpenGL': '/System/Library/Frameworks/OpenGL.framework/OpenGL',
            }
            path = frameworks.get(name)

        if path:
            lib = ctypes.cdll.LoadLibrary(path)
            if _debug_lib:
                print(path)
            if _debug_trace:
                lib = _TraceLibrary(lib)
            return lib

        msg = f"Can't find framework {name}."
        raise ImportError(msg)


class LinuxLibraryLoader(LibraryLoader):  # noqa: D101
    _ld_so_cache = None
    _local_libs_cache = None

    @staticmethod
    def _find_libs(directories: list[str]) -> dict[str, str]:
        libs = {}
        lib_re = re.compile(r'lib(.*)\.so(?:$|\.)')
        for directory in directories:
            try:
                for file in os.listdir(directory):
                    match = lib_re.match(file)
                    if match:
                        # Index by filename
                        path = os.path.join(directory, file)
                        if file not in libs:
                            libs[file] = path
                        # Index by library name
                        library = match.group(1)
                        if library not in libs:
                            libs[library] = path
            except OSError:  # noqa: PERF203
                pass
        return libs

    def _create_ld_so_cache(self) -> None:
        # Recreate search path followed by ld.so.  This is going to be
        # slow to build, and incorrect (ld.so uses ld.so.cache, which may
        # not be up-to-date).  Used only as fallback for distros without
        # /sbin/ldconfig.
        #
        # We assume the DT_RPATH and DT_RUNPATH binary sections are omitted.

        directories = []
        with contextlib.suppress(KeyError):
            directories.extend(os.environ['LD_LIBRARY_PATH'].split(':'))

        with contextlib.suppress(OSError), open('/etc/ld.so.conf') as fid:
            directories.extend([directory.strip() for directory in fid])

        directories.extend(['/lib', '/usr/lib'])

        self._ld_so_cache = self._find_libs(directories)

    def find_library(self, path: str) -> str:

        # search first for local libs
        if _local_lib_paths:
            if not self._local_libs_cache:
                self._local_libs_cache = self._find_libs(_local_lib_paths)
            if path in self._local_libs_cache:
                return self._local_libs_cache[path]

        # ctypes tries ldconfig, gcc and objdump.  If none of these are
        # present, we implement the ld-linux.so search path as described in
        # the man page.

        result = ctypes.util.find_library(path)

        if result:
            return result

        if self._ld_so_cache is None:
            self._create_ld_so_cache()

        return self._ld_so_cache.get(path)


if pyglet.compat_platform == 'darwin':
    loader = MacOSLibraryLoader()
elif pyglet.compat_platform.startswith('linux'):
    loader = LinuxLibraryLoader()
else:
    loader = LibraryLoader()

load_library = loader.load_library



# from ctypes import *
#
# from . import xkbcommon
#
#
# ctx = xkbcommon.xkb_context_new(xkbcommon.XKB_CONTEXT_NO_FLAGS)
# print(ctx)
#
# # TODO: get a real string from Wayland:
# keymap_string = create_string_buffer(b"ACTUAL KEYMAP STRING HERE")
#
# keymap = xkbcommon.xkb_keymap_new_from_string(ctx, keymap_string, xkbcommon.XKB_KEYMAP_FORMAT_TEXT_V1,
#                                               xkbcommon.XKB_KEYMAP_COMPILE_NO_FLAGS)
# print(keymap)
#
# state = xkbcommon.xkb_state_new(keymap)

import os

from pyglet.libs.wayland.gbm import *
from pyglet.libs.egl.egl import *
from pyglet.libs.egl.eglext import *

assert eglBindAPI(EGL_OPENGL_API) == EGL_TRUE

try:
    cards = [f for f in os.listdir('/dev/dri/') if f.startswith('card')]
    cards = [os.path.join('/dev/dri/', name) for name in cards]
    assert cards
except (OSError, AssertionError):
    print("unable to find DRI device")
    exit()

# Open the first GBM device, and create a surface
fd = os.open(cards[0], os.O_RDWR | os.MFD_CLOEXEC)
gbm_device = gbm_create_device(fd)
gbm_surface = gbm_surface_create(gbm_device, 256, 256, GBM_FORMAT_XRGB8888, GBM_BO_USE_SCANOUT | GBM_BO_USE_RENDERING)


# Get a native GBM platform display
dpy = eglGetPlatformDisplay(EGL_PLATFORM_GBM_MESA, gbm_device, None)
print("Display: ", dpy)
# Initialize EGL:
major, minor = EGLint(), EGLint()
result = eglInitialize(dpy, major, minor)
assert result == 1, "EGL Initialization Failed"
print("Initialized EGL: ", major.value, minor.value)

# Get the number of configs:
num_configs = EGLint()
result = eglGetConfigs(dpy, None, 0, num_configs)
assert result == EGL_TRUE, "Failed to query Configs"
print("Number of configs available: ", num_configs.value)

configs = (EGLConfig * num_configs.value)()
result = eglGetConfigs(dpy, configs, num_configs, byref(num_configs))
assert result == EGL_TRUE, "Failed to query Configs"

for config in configs:
    # For a test, just go with the first successful Config
    srf = eglCreatePlatformWindowSurface(dpy, config, gbm_surface, None)
    if srf:
        print("Surface: ", srf)
        break
else:
    print("No Surface Found")
    exit()


ctx = eglCreateContext(dpy, config, None, None)
print("Context:", ctx)

assert eglMakeCurrent(dpy, srf, srf, ctx) == EGL_TRUE, "Failed to Make Context Current"

eglDestroySurface(dpy, srf)
eglDestroyContext(dpy, ctx)
eglTerminate(dpy)

gbm_device_destroy(gbm_device)
os.close(fd)



#!/usr/bin/env python
"""Simple example of video playback.

Usage::

    video.py <filename>

See the Programming Guide for a partial list of supported video formats.
"""
import sys
import pyglet

if len(sys.argv) < 2:
    print(__doc__)
    sys.exit(1)

source = pyglet.media.load(sys.argv[1])
fmt = source.video_format
if not fmt:
    print('No video track in this source.')
    sys.exit(1)

player = pyglet.media.Player()
player.queue(source)
player.play()

window = pyglet.window.Window(width=fmt.width, height=fmt.height)


@window.event
def on_draw():
    player.texture.blit(0, 0)


pyglet.app.run()



#!/usr/bin/env python
"""Demonstrates how to handle a platform-specific event not defined in
pyglet by subclassing Window.  This is not for the faint-hearted!

A message will be printed to stdout when the following events are caught:

 - On Mac OS X, the window drag region is clicked.
 - On Windows, the display resolution is changed.
 - On Linux, the window properties are changed.

"""


import pyglet


_have_cocoa = _have_win32 = _have_xlib = False

if pyglet.compat_platform.startswith('linux'):
    _have_xlib = True
    from pyglet.window.xlib import XlibEventHandler, xlib

elif pyglet.compat_platform == 'darwin':
    _have_cocoa = True
    # from pyglet.window.cocoa import *

elif pyglet.compat_platform in ('win32', 'cygwin'):
    _have_win32 = True
    from pyglet.window.win32 import Win32EventHandler, WM_DISPLAYCHANGE


# Subclass Window
class MyWindow(pyglet.window.Window):
    if _have_cocoa:
        # TODO This is currently not supported in Cocoa (#156).
        pass

    if _have_win32:
        @Win32EventHandler(WM_DISPLAYCHANGE)
        def _on_window_display_change(self, msg, lParam, wParam):
            print('Display resolution changed.')
            return 0

    if _have_xlib:
        @XlibEventHandler(xlib.PropertyNotify)
        def _on_window_property_notify(self, event):
            print('Property notify.')


if __name__ == '__main__':
    window = MyWindow()
    pyglet.app.run()



"""Experimental multithreaded Job system.

This module provides the `JobExecutor` class, which allows submitting
batches of highly parallizable work in the form of functions and arguments.
This is commonly known as "task-based multithreading" or a "multithreaded
job system".

..note:: This module is only really useful with the recent Python 3.13
         releases that are built with experimental free-threading support.
         With typical Python releases that contain a GIL, this is not of
         much practical use.
"""

from __future__ import annotations

import os

from queue import Queue
from threading import Event, Thread
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Callable, Any


class _Worker(Thread):
    """A worker thread to pop and run jobs."""
    def __init__(self, workqueue: Queue, exitevent: Event, index: int) -> None:
        super().__init__(daemon=True)
        self._queue = workqueue
        self._exit = exitevent
        self._index = index
        self.start()

    def run(self) -> None:
        """Parallel thread of execution."""
        _exit = self._exit
        _queue = self._queue
        while not _exit.is_set():
            func, args = _queue.get()
            func(*args)
            _queue.task_done()

    def __repr__(self) -> str:
        return f"{self.name}(id={self.native_id})"


class JobExecutor:
    """A light-weight multithreaded job system.

    A JobExecutor loosely mimics the design of the Executor classes from the
    `concurrent.futures` module, but does not share any code with those classes.
    Instead, it is a more light-weight implementation intended for execution of
    highly parallizable functions (jobs). A key difference is that JobExecutor
    does not return `Futures`. Submitted jobs must be self-contained, or handle
    returning values by other mechanisms. The :py:meth:`~sync` method can be
    used to wait for all currently submitted jobs to complete.
    """

    def __init__(self, max_workers: int | None = None) -> None:
        """Create an instance of a JobExecutor.

        Args:
            max_workers: The number of threads to use. If `None`, will
                         create half as many threads as reported CPU cores.
        """
        self._max_workers = max_workers or os.cpu_count() // 2
        self._exitevent = Event()
        self._queue = Queue()
        self._threads = [_Worker(self._queue, self._exitevent, i) for i in range(self._max_workers)]

    def submit(self, func: Callable, *args: Any) -> None:
        """Submit a job to be executed on .

        A "job" consists of a function to be called, alone with any arguments
        that should be passed to it. Jobs are automatically executed by the next
        free worker thread. No values are returned when jobs are submitted,
        and the functions should also not return any values; return values are
        lost.

        Args:
            func: A function to execute.
            *args: Any arguments to pass to the function.
        """
        try:
            self._queue.put((func, args))
        except AttributeError:
            raise RuntimeError("cannot submit new tasks after shutdown")

    def shutdown(self) -> None:
        """Shut down the JobExecutor, terminating all worker threads.

        All JobExecutor workers are Daemon Threads, so it is not strictly
        necessary to call shutdown() at program termination. However, if
        it is no longer needed, shutdown() can be called to free up the
        thread resources.
        """
        if not self._queue:
            return
        self._exitevent.set()
        for _ in range(self._max_workers):
            self.submit(lambda: None)
        self._threads.clear()
        self._exitevent = None
        self._queue = None

    def sync(self) -> None:
        """Wait for all currently submitted jobs to complete.

        This method will wait until the internal queue is empty, AND all
        currently submitted jobs have completed execution. It can be used
        to fence between separate batches of jobs that should not be run
        at the same time. For example::

            for chunk in worklist:
                executor.submit(some_function, chunk)

            executor.sync()

            window.draw()

        """
        self._queue.join()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(workers={self._threads})"



"""OpenGL interface.

This package imports all OpenGL and registered OpenGL extension
functions.  Functions have identical signatures to their C counterparts.

OpenGL is documented in full at the `OpenGL Reference Pages`_.

The `OpenGL Programming Guide`_, also known as "The Red Book", is a popular
reference manual organised by topic. It is available in digital and paper
editions.

.. _OpenGL Reference Pages: https://www.khronos.org/registry/OpenGL-Refpages/
.. _OpenGL Programming Guide: http://opengl-redbook.com/

The following subpackages are imported into this "mega" package already
(and so are available by importing ``pyglet.gl``):

``pyglet.gl.gl``
    OpenGL
``pyglet.gl.gl.glext_arb``
    ARB registered OpenGL extension functions
``pyglet.gl.gl.gl_compat``
    Deprecated OpenGL extension functions.

These subpackages are also available, but are not imported into this namespace
by default:

``pyglet.gl.glext_nv``
    nVidia OpenGL extension functions
``pyglet.gl.agl``
    AGL (Mac OS X OpenGL context functions)
``pyglet.gl.glx``
    GLX (Linux OpenGL context functions)
``pyglet.gl.glxext_arb``
    ARB registered GLX extension functions
``pyglet.gl.glxext_nv``
    nvidia GLX extension functions
``pyglet.gl.wgl``
    WGL (Windows OpenGL context functions)
``pyglet.gl.wglext_arb``
    ARB registered WGL extension functions
``pyglet.gl.wglext_nv``
    nvidia WGL extension functions

The information modules are provided for convenience, and are documented below.
"""
from __future__ import annotations

import sys as _sys

import pyglet as _pyglet
from pyglet import compat_platform
from pyglet.gl.gl import *  # Must always be imported before gl_info or bad things happen.  # noqa: F403
from pyglet.gl import gl_info  # noqa: F401

from pyglet.gl.gl_compat import GL_INTENSITY, GL_LUMINANCE
from pyglet.gl.lib import GLException  # noqa: F401
from .gl import __all__ as _gl_all

from .base import DisplayConfig, Context, ObjectSpace  # noqa: F401, TCH001
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyglet.window import Window

_is_pyglet_doc_run = hasattr(_sys, "is_pyglet_doc_run") and _sys.is_pyglet_doc_run

#: The active OpenGL context.
#:
#: You can change the current context by calling `Context.set_current`;
#: do not modify this global.
#:
#: .. versionadded:: 1.1
current_context: Context | None = None


class ContextException(Exception):
    pass


class ConfigException(Exception):
    pass


if _pyglet.options['debug_texture']:
    _debug_texture_total = 0
    _debug_texture_sizes = {}
    _debug_texture = None


    def _debug_texture_alloc(texture, size):
        global _debug_texture_total

        _debug_texture_sizes[texture] = size
        _debug_texture_total += size

        print(f'{_debug_texture_total} (+{size})')


    def _debug_texture_dealloc(texture):
        global _debug_texture_total

        size = _debug_texture_sizes[texture]
        del _debug_texture_sizes[texture]
        _debug_texture_total -= size

        print(f'{_debug_texture_total} (-{size})')


    _glBindTexture = glBindTexture


    def glBindTexture(target, texture):
        global _debug_texture
        _debug_texture = texture
        return _glBindTexture(target, texture)


    _glTexImage2D = glTexImage2D


    def glTexImage2D(target, level, internalformat, width, height, border,
                     format, type, pixels):
        try:
            _debug_texture_dealloc(_debug_texture)
        except KeyError:
            pass

        if internalformat in (1, GL_ALPHA, GL_INTENSITY, GL_LUMINANCE):
            depth = 1
        elif internalformat in (2, GL_RGB16, GL_RGBA16):
            depth = 2
        elif internalformat in (3, GL_RGB):
            depth = 3
        else:
            depth = 4  # Pretty crap assumption
        size = (width + 2 * border) * (height + 2 * border) * depth
        _debug_texture_alloc(_debug_texture, size)

        return _glTexImage2D(target, level, internalformat, width, height, border, format, type, pixels)


    _glDeleteTextures = glDeleteTextures


    def glDeleteTextures(n, textures):
        if not hasattr(textures, '__len__'):
            _debug_texture_dealloc(textures.value)
        else:
            for i in range(n):
                _debug_texture_dealloc(textures[i].value)

        return _glDeleteTextures(n, textures)


def _create_shadow_window() -> None:
    global _shadow_window  # noqa: PLW0603

    import pyglet
    if not pyglet.options['shadow_window'] or _is_pyglet_doc_run:
        return

    from pyglet.window import Window

    class ShadowWindow(Window):
        _shadow = True
        def __init__(self) -> None:
            super().__init__(width=1, height=1, visible=False)

        def _create_projection(self) -> None:
            """Shadow window does not need a projection."""

        def _on_internal_resize(self, width: int, height: int) -> None:
            """No projection and not required."""

        def _on_internal_scale(self, scale: float, dpi: int) -> None:
            """No projection and not required."""

    _shadow_window = ShadowWindow()
    _shadow_window.switch_to()

    from pyglet import app
    app.windows.remove(_shadow_window)


if _is_pyglet_doc_run:
    from .base import Config

elif _pyglet.options['headless']:
    from .headless import HeadlessConfig as Config
elif compat_platform in ('win32', 'cygwin'):
    from .win32 import Win32Config as Config
elif compat_platform.startswith('linux'):
    from .xlib import XlibConfig as Config
elif compat_platform == 'darwin':
    from .cocoa import CocoaConfig as Config


_shadow_window: Window | None = None

# Import pyglet.window now if it isn't currently being imported (this creates the shadow window).
if not _is_pyglet_doc_run and 'pyglet.window' not in _sys.modules and _pyglet.options['shadow_window']:
    # trickery is for circular import
    _pyglet.gl = _sys.modules[__name__]
    import pyglet.window  # noqa: F401

__all__ = ['Config', 'GLException', 'gl_info']
__all__.extend(_gl_all)



import pyglet
from game import load, player

# Set up a window
game_window = pyglet.window.Window(800, 600)

main_batch = pyglet.graphics.Batch()

# Set up the two top labels
score_label = pyglet.text.Label(text="Score: 0", x=10, y=575, batch=main_batch)
level_label = pyglet.text.Label(text="Version 4: Bullets and Structure",
                                x=400, y=575, anchor_x='center', batch=main_batch)

# Initialize the player sprite
player_ship = player.Player(x=400, y=300, batch=main_batch)

# Make three sprites to represent remaining lives
player_lives = load.player_lives(2, main_batch)

# Make three asteroids so we have something to shoot at
asteroids = load.asteroids(3, player_ship.position, main_batch)

# Store all objects that update each frame in a list
game_objects = [player_ship] + asteroids

# Add any specified event handlers to the event handler stack
for obj in game_objects:
    for handler in obj.event_handlers:
        game_window.push_handlers(handler)


@game_window.event
def on_draw():
    game_window.clear()
    main_batch.draw()


def update(dt):
    # To avoid handling collisions twice, we employ nested loops of ranges.
    # This method also avoids the problem of colliding an object with itself.
    for i in range(len(game_objects)):
        for j in range(i + 1, len(game_objects)):

            obj_1 = game_objects[i]
            obj_2 = game_objects[j]

            # Make sure the objects haven't already been killed
            if not obj_1.dead and not obj_2.dead:
                if obj_1.collides_with(obj_2):
                    obj_1.handle_collision_with(obj_2)
                    obj_2.handle_collision_with(obj_1)

    # Let's not modify the list while traversing it
    to_add = []

    for obj in game_objects:
        obj.update(dt)
        to_add.extend(obj.new_objects)
        obj.new_objects = []

    # Get rid of dead objects
    for to_remove in [obj for obj in game_objects if obj.dead]:
        # If the dying object spawned any new objects, add those to the game_objects list later
        to_add.extend(obj.new_objects)

        # Remove the object from any batches it is a member of
        to_remove.delete()

        # Remove the object from our list
        game_objects.remove(to_remove)

    # Add new objects to the list
    game_objects.extend(to_add)


if __name__ == "__main__":
    # Update the game 120 times per second
    pyglet.clock.schedule_interval(update, 1 / 120.0)

    # Tell pyglet to do its thing
    pyglet.app.run()



import pytest

from tests.base.interactive import InteractiveTestCase

from pyglet import app
from pyglet import gl
from pyglet import graphics
from pyglet import text
from pyglet.text import caret
from pyglet.text import layout
from pyglet import window
from pyglet.window import key


class TestWindow(window.Window):

    INSTRUCTIONS =\
        "Please observe whether the caret color matches" \
        "{0}, then close the window."

    def __init__(
        self,
        *args,
        caret_color=(0, 0, 0, 255),
        **kwargs
    ):
        super().__init__(*args, **kwargs)

        self.batch = graphics.Batch()
        self.document = text.decode_text(
            self.INSTRUCTIONS.format(caret_color)
        )

        self.margin = 2
        self.layout = layout.IncrementalTextLayout(
            self.document,
            width=self.width - self.margin * 2,
            height=self.height - self.margin * 2,
            multiline=True,
            batch=self.batch)
        self.caret = caret.Caret(self.layout, color=caret_color)
        self.push_handlers(self.caret)

        self.set_mouse_cursor(self.get_system_mouse_cursor('text'))

    def on_resize(self, width, height):
        super().on_resize(width, height)
        self.layout.begin_update()
        self.layout.x = self.margin
        self.layout.y = self.margin
        self.layout.width = width - self.margin * 2
        self.layout.height = height - self.margin * 2
        self.layout.end_update()

    def on_mouse_scroll(self, x, y, scroll_x, scroll_y):
        self.layout.view_x -= scroll_x
        self.layout.view_y += scroll_y * 16

    def on_draw(self):
        gl.glClearColor(1, 1, 1, 1)
        self.clear()
        self.batch.draw()

    def on_key_press(self, symbol, modifiers):
        super().on_key_press(symbol, modifiers)
        if symbol == key.TAB:
            self.caret.on_text('\t')


class _DRYHelperMixin:

    def build_window(self, color=(0, 0, 0, 255), window_class=TestWindow):
        self.window = window_class(caret_color=color, resizable=False, visible=False)
        self.window.set_visible()

    def build_window_via_setter(self, color, window_class=TestWindow):
        self.build_window(window_class=window_class)
        self.window.caret.color = color

    def ask_color(self, color):
        self.user_verify(f"Was the caret color {color}?", take_screenshot=False)


@pytest.mark.requires_user_action
class CaretColorInitTestCase(InteractiveTestCase, _DRYHelperMixin):
    """Test if the text area Caret color is set correctly by init

    Examine & remember the color of the caret, and type Y + enter if the
    caret color matched the numbers in the label.

    Press ESC to exit the test.
    """

    def test_caret_color_init_rgb(self):
        color = (255, 0, 0)
        self.build_window(color)
        app.run()
        self.ask_color(color)

    def test_caret_color_init_rgba(self):
        # 80 seems to be an ok balance between visibility and
        # ease of telling apart from alpha == 255.
        color = (255, 0, 0, 80)
        self.build_window(color)
        app.run()
        self.ask_color(color)


class SetterTestWindow(TestWindow):

    INSTRUCTIONS = \
        "Press the tab key & observe whether the caret color" \
        "changed to {0}, then close the window."

    def on_key_press(self, symbol, modifiers):
        super().on_key_press(symbol, modifiers)
        if symbol == key.TAB:
            self.caret.color = (0, 255, 0, 10)


@pytest.mark.requires_user_action
class CaretColorSetterTestCase(InteractiveTestCase, _DRYHelperMixin):
    """Test if the Caret's color setter works correctly

    Examine & remember the color of the caret, and type Y + enter if the
    caret color matched the numbers in the label.

    Press ESC to exit the test.
    """

    def test_caret_color_setter_rgb(self):
        color = (0, 0, 255)
        self.build_window_via_setter(color, window_class=SetterTestWindow)
        app.run()
        self.ask_color(color)

    def test_caret_color_setter_rgba(self):
        # 80 seems to be an ok balance between visibility and
        # ease of telling apart from alpha == 255.
        color = (0, 0, 255, 0)
        self.build_window_via_setter(color, window_class=SetterTestWindow)
        app.run()
        self.ask_color(color)




"""Display different types of interactive widgets."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pyglet
from pyglet.event import EventDispatcher
from pyglet.graphics import Group
from pyglet.text.caret import Caret
from pyglet.text.layout import IncrementalTextLayout

if TYPE_CHECKING:
    from pyglet.graphics import Batch
    from pyglet.image import AbstractImage


class WidgetBase(EventDispatcher):
    """The base of all widgets."""
    def __init__(self, x: int, y: int, width: int, height: int) -> None:  # noqa: D107
        self._x = x
        self._y = y
        self._width = width
        self._height = height
        self._parent = None
        self._bg_group = None
        self._fg_group = None
        self._enabled = True

    def _set_enabled(self, enabled: bool) -> None:
        """Internal hook for setting enabled.

        Override this in subclasses to perform effects when a widget is
        enabled or disabled.
        """

    @property
    def enabled(self) -> bool:
        """Get/set whether this widget is enabled.

        To react to changes in this value, override
        :py:meth:`._set_enabled` on widgets. For example, you may want
        to cue the user by:

        - Playing an animation and/or sound
        - Setting a highlight color
        - Displaying a toast or notification
        """
        return self._enabled

    @enabled.setter
    def enabled(self, new_enabled: bool) -> None:
        if self._enabled == new_enabled:
            return
        self._enabled = new_enabled
        self._set_enabled(new_enabled)

    def update_groups(self, order: int) -> None:
        pass

    @property
    def x(self) -> int:
        """X coordinate of the widget."""
        return self._x

    @x.setter
    def x(self, value: int) -> None:
        self._x = value
        self._update_position()
        self.dispatch_event("on_reposition", self)

    @property
    def y(self) -> int:
        """Y coordinate of the widget."""
        return self._y

    @y.setter
    def y(self, value: int) -> None:
        self._y = value
        self._update_position()
        self.dispatch_event("on_reposition", self)

    @property
    def parent(self):
        """The frame this widget belongs to.

        :type: `~pyglet.gui.frame.Frame`
        """
        return self._parent

    @parent.setter
    def parent(self, value):
        self._parent = value

    @property
    def position(self) -> tuple[int, int]:
        """The x, y coordinate of the widget as a tuple."""
        return self._x, self._y

    @position.setter
    def position(self, values: tuple[int, int]) -> None:
        self._x, self._y = values
        self._update_position()
        self.dispatch_event("on_reposition", self)

    @property
    def width(self) -> int:
        """Width of the widget."""
        return self._width

    @property
    def height(self) -> int:
        """Height of the widget."""
        return self._height

    @property
    def aabb(self) -> tuple[int, int, int, int]:
        """Bounding box of the widget.

        The "left", "bottom", "right", and "top" coordinates of the
        widget. This is expressed as (x, y, x + width, y + height).
        """
        return self._x, self._y, self._x + self._width, self._y + self._height

    @property
    def value(self) -> int | float | bool:
        """Query or set the Widget's value.

        This property allows you to set the value of a Widget directly, without any
        user input.  This could be used, for example, to restore Widgets to a
        previous state, or if some event in your program is meant to naturally
        change the same value that the Widget controls.  Note that events are not
        dispatched when changing this property directly.
        """
        raise NotImplementedError('Value depends on control type!')

    @value.setter
    def value(self, value: int | float | bool) -> None:
        raise NotImplementedError('Value depends on control type!')

    def _check_hit(self, x: int, y: int) -> bool:
        return self._x < x < self._x + self._width and self._y < y < self._y + self._height

    def _update_position(self) -> None:
        raise NotImplementedError('Unable to reposition this Widget')

    # Handlers

    def on_key_press(self, symbol: int, modifiers: int) -> None:
        pass

    def on_key_release(self, symbol: int, modifiers: int) -> None:
        pass

    def on_mouse_press(self, x: int, y: int, buttons: int, modifiers: int) -> None:
        pass

    def on_mouse_release(self, x: int, y: int, buttons: int, modifiers: int) -> None:
        pass

    def on_mouse_drag(self, x: int, y: int, dx: int, dy: int, buttons: int, modifiers: int) -> None:
        pass

    def on_mouse_scroll(self, x: int, y: int, scroll_x: float, scroll_y: float) -> None:
        pass

    def on_mouse_motion(self, x: int, y: int, dx: int, dy: int) -> None:
        pass

    def on_mouse_enter(self, x: int, y: int) -> None:
        pass

    def on_mouse_leave(self, x: int, y: int) -> None:
        pass

    def on_text(self, text: str) -> None:
        pass

    def on_text_motion(self, motion: int) -> None:
        pass

    def on_text_motion_select(self, motion: int) -> None:
        pass


WidgetBase.register_event_type("on_reposition")


class PushButton(WidgetBase):
    """Instance of a push button.

    Triggers the event 'on_press' when it is clicked by the mouse.
    Triggers the event 'on_release' when the mouse is released.
    """

    def __init__(self, x: int, y: int,
                 pressed: AbstractImage,
                 unpressed: AbstractImage,
                 hover: AbstractImage | None = None,
                 batch: Batch | None = None,
                 group: Group | None = None) -> None:
        """Create a push button.

        Args:
            x:
                X coordinate of the push button.
            y:
                Y coordinate of the push button.
            pressed:
                Image to display when the button is pressed.
            unpressed:
                Image to display when the button isn't pressed.
            hover:
                Image to display when the button is being hovered over.
            batch:
                Optional batch to add the push button to.
            group:
                Optional parent group of the push button.
        """
        super().__init__(x, y, unpressed.width, unpressed.height)
        self._pressed_img = pressed
        self._unpressed_img = unpressed
        self._hover_img = hover or unpressed

        self._batch = batch or pyglet.graphics.Batch()
        self._user_group = group
        bg_group = Group(order=0, parent=group)
        self._sprite = pyglet.sprite.Sprite(self._unpressed_img, x=x, y=y, batch=batch, group=bg_group)

        self._pressed = False

    def _update_position(self) -> None:
        self._sprite.position = self._x, self._y, 0

    @property
    def value(self) -> bool:
        """Whether the button is pressed or not."""
        return self._pressed

    @value.setter
    def value(self, value: bool) -> None:
        assert type(value) is bool, "This Widget's value must be True or False."
        self._pressed = value
        self._sprite.image = self._pressed_img if self._pressed else self._unpressed_img

    def update_groups(self, order: int) -> None:
        self._sprite.group = Group(order=order + 1, parent=self._user_group)

    def on_mouse_press(self, x: int, y: int, buttons: int, modifiers: int) -> None:
        if not self.enabled or not self._check_hit(x, y):
            return
        self._sprite.image = self._pressed_img
        self._pressed = True
        self.dispatch_event('on_press', self)

    def on_mouse_release(self, x: int, y: int, buttons: int, modifiers: int) -> None:
        if not self.enabled or not self._pressed:
            return
        self._sprite.image = self._hover_img if self._check_hit(x, y) else self._unpressed_img
        self._pressed = False
        self.dispatch_event('on_release', self)

    def on_mouse_leave(self, x: int, y: int) -> None:
        if not self.enabled or not self._pressed:
            return
        self._sprite.image = self._unpressed_img
        self._pressed = False
        self.dispatch_event('on_release')

    def on_mouse_motion(self, x: int, y: int, dx: int, dy: int) -> None:
        if not self.enabled or self._pressed:
            return
        self._sprite.image = self._hover_img if self._check_hit(x, y) else self._unpressed_img

    def on_mouse_drag(self, x: int, y: int, dx: int, dy: int, buttons: int, modifiers: int) -> None:
        if not self.enabled or self._pressed:
            return
        self._sprite.image = self._hover_img if self._check_hit(x, y) else self._unpressed_img

    def on_press(self, widget: PushButton) -> None:
        """Event: Dispatched when the button is clicked."""

    def on_release(self, widget: PushButton) -> None:
        """Event: Dispatched when the button is released."""


PushButton.register_event_type('on_press')
PushButton.register_event_type('on_release')


class ToggleButton(PushButton):
    """Instance of a toggle button.

    Triggers the event 'on_toggle' when the mouse is pressed or released.
    """

    def _get_release_image(self, x: int, y: int) -> AbstractImage:
        return self._hover_img if self._check_hit(x, y) else self._unpressed_img

    def on_mouse_press(self, x: int, y: int, buttons: int, modifiers: int) -> None:
        if not self.enabled or not self._check_hit(x, y):
            return
        self._pressed = not self._pressed
        self._sprite.image = self._pressed_img if self._pressed else self._get_release_image(x, y)
        self.dispatch_event('on_toggle', self, self._pressed)

    def on_mouse_release(self, x: int, y: int, buttons: int, modifiers: int) -> None:
        if not self.enabled or self._pressed:
            return
        self._sprite.image = self._get_release_image(x, y)

    def on_toggle(self, widget: ToggleButton, value: bool) -> None:
        """Event: returns True or False to indicate the current state."""


ToggleButton.register_event_type('on_toggle')


class Slider(WidgetBase):
    """Instance of a slider made of a base and a knob image.

    Triggers the event 'on_change' when the knob position is changed.
    The knob position can be changed by dragging with the mouse, or
    scrolling the mouse wheel.
    """

    def __init__(self, x: int, y: int,
                 base: AbstractImage, knob: AbstractImage,
                 edge: int = 0,
                 batch: Batch | None = None,
                 group: Group | None = None) -> None:
        """Create a slider.

        Args:
            x:
                X coordinate of the slider.
            y:
                Y coordinate of the slider.
            base:
                Image to display as the background to the slider.
            knob:
                Knob that moves to show the position of the slider.
            edge:
                Pixels from the maximum and minimum position of the slider,
                to the edge of the base image.
            batch:
                Optional batch to add the slider to.
            group:
                Optional parent group of the slider.
        """
        super().__init__(x, y, base.width, knob.height)
        self._edge = edge
        self._base_img = base
        self._knob_img = knob
        self._half_knob_width = knob.width / 2
        self._half_knob_height = knob.height / 2
        self._knob_img.anchor_y = int(knob.height / 2)

        self._min_knob_x = x + edge
        self._max_knob_x = x + base.width - knob.width - edge

        self._user_group = group
        bg_group = Group(order=0, parent=group)
        fg_group = Group(order=1, parent=group)
        self._base_spr = pyglet.sprite.Sprite(self._base_img, x, y, batch=batch, group=bg_group)
        self._knob_spr = pyglet.sprite.Sprite(self._knob_img, x+edge, y+base.height/2, batch=batch, group=fg_group)

        self._value = 0
        self._in_update = False

    def _update_position(self) -> None:
        self._base_spr.position = self._x, self._y, 0
        self._min_knob_x = self._x + self._edge
        self._max_knob_x = self._x + self._base_img.width - self._knob_img.width - self._edge
        x = (self._max_knob_x - self._min_knob_x) * self._value / 100 + self._min_knob_x + self._half_knob_width
        x = max(self._min_knob_x, min(x - self._half_knob_width, self._max_knob_x))
        y = self._y + self._base_img.height / 2
        self._knob_spr.position = x, y, 0

    @property
    def value(self) -> float:
        return self._value

    @value.setter
    def value(self, value: float) -> None:
        assert type(value) in (int, float), "This Widget's value must be an int or float."
        self._value = value
        x = (self._max_knob_x - self._min_knob_x) * value / 100 + self._min_knob_x + self._half_knob_width
        self._knob_spr.x = max(self._min_knob_x, min(x - self._half_knob_width, self._max_knob_x))

    def update_groups(self, order: int) -> None:
        self._base_spr.group = Group(order=order + 1, parent=self._user_group)
        self._knob_spr.group = Group(order=order + 2, parent=self._user_group)

    @property
    def _min_x(self) -> int:
        return self._x + self._edge

    @property
    def _max_x(self) -> int:
        return self._x + self._width - self._edge

    @property
    def _min_y(self) -> int:
        return int(self._y - self._half_knob_height)

    @property
    def _max_y(self) -> int:
        return int(self._y + self._half_knob_height + self._base_img.height / 2)

    def _check_hit(self, x: int, y: int) -> bool:
        return self._min_x < x < self._max_x and self._min_y < y < self._max_y

    def _update_knob(self, x: int) -> None:
        self._knob_spr.x = max(self._min_knob_x, min(x - self._half_knob_width, self._max_knob_x))
        self._value = abs(((self._knob_spr.x - self._min_knob_x) * 100) / (self._min_knob_x - self._max_knob_x))
        self.dispatch_event('on_change', self, self._value)

    def on_mouse_press(self, x: int, y: int, buttons: int, modifiers: int) -> None:
        if not self.enabled:
            return
        if self._check_hit(x, y):
            self._in_update = True
            self._update_knob(x)

    def on_mouse_drag(self, x: int, y: int, dx: int, dy: int, buttons: int, modifiers: int) -> None:
        if not self.enabled:
            return
        if self._in_update:
            self._update_knob(x)

    def on_mouse_scroll(self, x: int, y: int, scroll_x: float, scroll_y: float) -> None:
        if not self.enabled:
            return
        if self._check_hit(x, y):
            self._update_knob(self._knob_spr.x + self._half_knob_width + scroll_y)  # type: ignore reportArgumentType

    def on_mouse_release(self, x: int, y: int, buttons: int, modifiers: int) -> None:
        if not self.enabled:
            return
        self._in_update = False

    def on_change(self, widget: Slider, value: float) -> None:
        """Event: Returns the current value when the slider is changed."""


Slider.register_event_type('on_change')


class TextEntry(WidgetBase):
    """Instance of a text entry widget. Allows the user to enter and submit text.

    Triggers the event 'on_commit', when the user hits the Enter or Return key.
    The current text string is passed along with the event.
    """

    def __init__(self, text: str,
                 x: int, y: int, width: int,
                 color: tuple[int, int, int, int] = (255, 255, 255, 255),
                 text_color: tuple[int, int, int, int] = (0, 0, 0, 255),
                 caret_color: tuple[int, int, int, int] = (0, 0, 0, 255),
                 batch: Batch | None = None,
                 group: Group | None = None) -> None:
        """Create a text entry widget.

        Args:
            text:
                Initial text to display.
            x:
                X coordinate of the text entry widget.
            y:
                Y coordinate of the text entry widget.
            width:
                The width of the text entry widget.
            color:
                The color of the outline box in RGBA format.
            text_color:
                The color of the text in RGBA format.
            caret_color:
                The color of the caret (when it is visible) in RGBA or RGB format.
            batch:
                Optional batch to add the text entry widget to.
            group:
                Optional parent group of text entry widget.
        """
        self._doc = pyglet.text.document.UnformattedDocument(text)
        self._doc.set_style(0, len(self._doc.text), {'color': text_color})
        font = self._doc.get_font()
        height = font.ascent - font.descent

        self._user_group = group
        bg_group = Group(order=0, parent=group)
        fg_group = Group(order=1, parent=group)

        # Rectangular outline with 2-pixel pad:
        self._pad = p = 2
        self._outline = pyglet.shapes.Rectangle(x-p, y-p, width+p+p, height+p+p, color, batch=batch, group=bg_group)

        # Text and Caret:
        self._layout = IncrementalTextLayout(self._doc, x, y, 0, width, height, batch=batch, group=fg_group)
        self._caret = Caret(self._layout, color=caret_color)
        self._caret.visible = False

        self._focus = False

        super().__init__(x, y, width, height)

    def _update_position(self) -> None:
        self._layout.position = self._x, self._y, 0
        self._outline.position = self._x - self._pad, self._y - self._pad

    @property
    def value(self) -> str:
        return self._doc.text

    @value.setter
    def value(self, value: str) -> None:
        assert type(value) is str, "This Widget's value must be a string."
        self._doc.text = value

    @property
    def width(self) -> int:
        return self._width

    @width.setter
    def width(self, value: int) -> None:
        self._width = value
        self._layout.width = value
        self._outline.width = value

    @property
    def height(self) -> int:
        return self._height

    @height.setter
    def height(self, value: int) -> None:
        self._height = value
        self._layout.height = value
        self._outline.height = value

    @property
    def focus(self) -> bool:
        return self._focus

    @focus.setter
    def focus(self, value: bool) -> None:
        self._set_focus(value)

    def _check_hit(self, x: int, y: int) -> bool:
        return self._x < x < self._x + self._width and self._y < y < self._y + self._height

    def _set_focus(self, value: bool) -> None:
        self._focus = value
        self._caret.visible = value
        self._caret.layout = self._layout

    def update_groups(self, order: int) -> None:
        self._outline.group = Group(order=order + 1, parent=self._user_group)
        self._layout.group = Group(order=order + 2, parent=self._user_group)

    def on_mouse_motion(self, x: int, y: int, dx: int, dy: int) -> None:
        if not self.enabled:
            return

    def on_mouse_drag(self, x: int, y: int, dx: int, dy: int, buttons: int, modifiers: int) -> None:
        if not self.enabled:
            return
        if self._focus:
            self._caret.on_mouse_drag(x, y, dx, dy, buttons, modifiers)

    def on_mouse_press(self, x: int, y: int, buttons: int, modifiers: int) -> None:
        if not self.enabled:
            return
        if self._check_hit(x, y):
            self._set_focus(True)
            self._caret.on_mouse_press(x, y, buttons, modifiers)
        else:
            self._set_focus(False)

    def on_text(self, text: str) -> None:
        if not self.enabled:
            return
        if self._focus:
            # Commit on Enter/Return:
            if text in ('\r', '\n'):
                self.dispatch_event('on_commit', self, self._layout.document.text)
                self._set_focus(False)
                return
            self._caret.on_text(text)

    def on_text_motion(self, motion: int) -> None:
        if not self.enabled:
            return
        if self._focus:
            self._caret.on_text_motion(motion)

    def on_text_motion_select(self, motion: int) -> None:
        if not self.enabled:
            return
        if self._focus:
            self._caret.on_text_motion_select(motion)

    def on_commit(self, widget: TextEntry, text: str) -> None:
        """Event: dispatches the current text when committed via Enter/Return key."""


TextEntry.register_event_type('on_commit')



"""Concept for higher level instance generation and organization."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Protocol, Sequence

if TYPE_CHECKING:
    from pyglet.graphics import Batch
    from pyglet.graphics.vertexdomain import VertexInstance, VertexList


class VertexListObject(Protocol):
    batch: Batch
    _vertex_list: VertexList


class InstanceSourceMixin:
    _vertex_list: VertexList

    def __init__(self, attributes: Sequence[str]) -> None:
        self._attributes = tuple(attributes)
        self._instances = []

    @property
    def instances(self) -> list[VertexInstance]:
        return self._instances

    def create_instance(self, **kwargs: Any) -> VertexInstance:
        instance = self._vertex_list.add_instance(**kwargs)
        self._instances.append(instance)
        return instance

    def pop(self) -> VertexInstance:
        instance = self._instances.pop()
        instance.delete()
        return instance


class InstanceGenerator:
    _attributes: tuple[str, ...]

    def __init__(self, source: VertexListObject, attributes: Sequence[str]) -> None:
        self.source = source
        self._attributes = tuple(attributes)
        self._instances = []

        if self.source._vertex_list is None:
            raise Exception("No vertex list")

        if self.source.batch is None:
            raise Exception("No batch")

        self.migrate()

    @property
    def instance_attributes(self) -> tuple[str, ...]:
        return self._attributes

    def migrate(self) -> None:
        vlist = self.source._vertex_list
        batch = self.source.batch

        instanced_domain = batch._convert_to_instanced(vlist.domain, self._attributes)
        vlist.set_instance_source(instanced_domain, self._attributes)

    @property
    def instances(self) -> list[VertexInstance]:
        return self._instances

    def create(self, **kwargs: Any) -> VertexInstance:
        instance = self.source._vertex_list.add_instance(**kwargs)
        self._instances.append(instance)
        return instance

    def pop(self) -> VertexInstance:
        instance = self._instances.pop()
        instance.delete()
        return instance



#!/usr/bin/python

import pyglet
pyglet.options["debug_input"] = True

window = pyglet.window.Window()
tablets = pyglet.input.get_tablets()
canvases = []

if tablets:
    print('Tablets:')
    for i, tablet in enumerate(tablets):
        print(f'  ({i + 1}) {tablet.name}')
    print('Press number key to open corresponding tablet device.')
else:
    print('No tablets found.')


@window.event
def on_text(text):
    try:
        index = int(text) - 1
    except ValueError:
        return

    if not (0 <= index < len(tablets)):
        return

    name = tablets[index].name

    try:
        canvas = tablets[index].open(window)
    except pyglet.input.DeviceException:
        print(f'Failed to open tablet {index} on window')
        return

    print(f'Opened {name}')

    @canvas.event
    def on_enter(cursor):
        print(f'{name}: on_enter({cursor!r})')


    @canvas.event
    def on_leave(cursor):
        print(f'{name}: on_leave({cursor!r})')

    @canvas.event
    def on_motion(cursor, x, y, pressure, tilt_x, tilt_y, buttons):
        print(f'{name}: on_motion({cursor!r}, {x!r}, {y!r}, {pressure!r}, {tilt_x!r}, {tilt_y!r}, {buttons!r})')

    # If ExpressKey is supported for the OS, the events will be supported.
    if 'on_express_key_press' in canvas.event_types:
        @canvas.event
        def on_express_key_press(control_id, location_id):
            print(f'on_express_key_press(control_id={control_id}, location_id={location_id}')

        @canvas.event
        def on_express_key_release(control_id, location_id):
            print(f'on_express_key_release(control_id={control_id}, location_id={location_id}')

@window.event
def on_mouse_press(x, y, button, modifiers):
    print(f'on_mouse_press({x!r}, {y!r}, {button!r}, {modifiers!r}')


@window.event
def on_mouse_release(x, y, button, modifiers):
    print(f'on_mouse_release{x!r}, {y!r}, {button!r}, {modifiers!r}')


pyglet.app.run()
